!===============================================================================
!
! flip_bc_schfun.F90  ---  a boundary condition for Schroedinger function
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine flip_bc1_schfun(u_flip, u)

  use module_schr_bc
  use module_lattice
  use module_vol
  implicit none

  integer, dimension(DIM) :: i_pe
  GAUGE_FIELD, intent(in)  :: u
  GAUGE_FIELD, intent(out) :: u_flip
  integer :: mu, nu, count, i, eo

  u_flip = u

  mu = gamma_index(4)
  do eo = EVEN,ODD

     !$omp parallel do private(i)
     do count = 1, bc_len_bb(eo)
        i = bc_list_bb(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_tt(eo)
        i = bc_list_tt(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_b(eo)
        i = bc_list_b(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_t(eo)
        i = bc_list_t(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

  enddo

end

!-------------------------------------------------------------------------------
subroutine flip_bc2_schfun(u_flip, u)

  use module_schr_bc
  use module_lattice
  use module_vol
  implicit none

  integer, dimension(DIM) :: i_pe
  GAUGE_FIELD, intent(in)  :: u
  GAUGE_FIELD, intent(out) :: u_flip
  integer :: mu, nu, count, i, eo

  u_flip = u

  mu = gamma_index(4)
  do eo = EVEN,ODD

     !$omp parallel do private(i)
     do count = 1, bc_len_bb(eo)
        i = bc_list_bb(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_tt(eo)
        i = bc_list_tt(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_b(eo)
        i = bc_list_b(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

     !$omp parallel do private(i)
     do count = 1, bc_len_t(eo)
        i = bc_list_t(count, eo)
        call u_zero(u_flip(1, 1, i, eo, mu))
     enddo

  enddo

end

!===============================================================================
