!===============================================================================
!
! init_schr_bc.F90 -- init schr boundary condition
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_schr_bc

  INTEGER, dimension(:, :), pointer, save :: bc_list_b
  INTEGER, dimension(:, :), pointer, save :: bc_list_t
  INTEGER, dimension(:, :), pointer, save :: bc_list_bb
  INTEGER, dimension(:, :), pointer, save :: bc_list_tt
  INTEGER, dimension(:, :), pointer, save :: bc_list_in

  integer, dimension(EVEN:ODD), save      :: bc_len_b
  integer, dimension(EVEN:ODD), save      :: bc_len_t
  integer, dimension(EVEN:ODD), save      :: bc_len_bb
  integer, dimension(EVEN:ODD), save      :: bc_len_tt
  integer, dimension(EVEN:ODD), save      :: bc_len_in

end

!-------------------------------------------------------------------------------
subroutine init_schr_bc()

  use      module_schr_bc
  use      module_function_decl
  use      module_lattice
  use      module_vol
  implicit none

  integer, dimension(DIM) :: i0, i1, i_pe, j
  integer :: me, mu, x, y, z, t, i, eo, count(EVEN:ODD)
  integer, external :: xyzt2i, e_o

  call my_coord(i_pe)
  mu = gamma_index(4)
  if (NPE(mu) == 1 ) call die("so far schr bc does not work in NPE(4)=1 ")

  allocate(bc_list_b(volh_tot, EVEN:ODD))
  allocate(bc_list_t(volh_tot, EVEN:ODD))
  allocate(bc_list_bb(volh_tot, EVEN:ODD))
  allocate(bc_list_tt(volh_tot, EVEN:ODD))
  allocate(bc_list_in(volh_tot, EVEN:ODD))
  bc_len_b = 0
  bc_len_t = 0
  bc_len_bb = 0
  bc_len_tt = 0
  bc_len_in = 0

!==========================================================t
!----------------------------------------------------------t
  if (i_pe(mu) == (NPE(mu) - 1)) then
     i0 = 0
     i1 = N - 1
     i0(mu) = N(mu) - 1
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_t(eo) = bc_len_t(eo) + 1
        bc_list_t(bc_len_t(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
!----------------------------------------------------------tt
     i0 = 0
     i1 = N - 1
     i0(mu) = N(mu)
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_tt(eo) = bc_len_tt(eo) + 1
        bc_list_tt(bc_len_tt(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
!----------------------------------------------------------t in
     i0 = 0
     i1 = N - 1
     i0(mu) = N(mu) -2
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_tt(eo) = bc_len_tt(eo) + 1
        bc_list_tt(bc_len_tt(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
!==========================================================b
!----------------------------------------------------------b
  elseif (i_pe(mu) == 0) then
     i0 = 0
     i1 = N - 1
     i0(mu) = 0
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_b(eo) = bc_len_b(eo) + 1
        bc_list_b(bc_len_b(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
!----------------------------------------------------------bb
     i0 = 0
     i1 = N - 1
     i0(mu) = -1
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_bb(eo) = bc_len_bb(eo) + 1
        bc_list_bb(bc_len_bb(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
!----------------------------------------------------------b in
     i0 = 0
     i1 = N - 1
     i0(mu) = 1
     i1(mu) = i0(mu)
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_in(eo) = bc_len_in(eo) + 1
        bc_list_in(bc_len_in(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
  else
!==========================================================in
     i0 = 0
     i1 = N - 1
     do t = i0(4), i1(4)
     do z = i0(3), i1(3)
     do y = i0(2), i1(2)
     do x = i0(1), i1(1)
        j = (/x, y, z, t/)
        i = xyzt2i(j)
        eo = e_o(j)
        bc_len_in(eo) = bc_len_in(eo) + 1
        bc_list_in(bc_len_in(eo), eo) = i
     enddo
     enddo
     enddo
     enddo
  endif

end

!===============================================================================
