!===============================================================================
!
! schr_boundary.F90  ---  a boundary condition for Schroedinger function
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

#ifdef QPACE
# undef PI
# define PI    STRCAT(3.1415926535897932384626_, RKIND)
#endif

!-------------------------------------------------------------------------------
subroutine schr_boundary_gauge(u)

  use module_schr_bc
  use module_lattice
  use module_vol
  use module_nn
  use module_function_decl
  implicit none

  GAUGE_FIELD, intent(inout) :: u
  integer :: mu, nu, count, i, eo, c1, c2, mu1
  COMPLEX :: bound, bound1(3), bound2(3)

REAL,external :: Re_Tr_uu,sg

SU3::uuu
integer::j1,j2,o,e
REAL::p


  bound1(1) = cmplx(dcos(PI/SIX), -dsin(PI/SIX), kind = RKIND)
  bound1(2) = cmplx(ONE, ZERO, kind = RKIND)
  bound1(3) = cmplx(dcos(PI/SIX), dsin(PI/SIX), kind = RKIND)
  bound2(1) = cmplx(dcos(5.0_8*PI/SIX), -dsin(5.0_8*PI/SIX), kind = RKIND)
  bound2(2) = cmplx(dcos(PI/THREE), dsin(PI/THREE), kind = RKIND)
  bound2(3) = cmplx(dcos(PI/TWO), dsin(PI/TWO), kind = RKIND)

write(STDERR,*)sg(u)
  do mu = 1, 1!4!3 !DIM
     mu1 = gamma_index(mu)
     do eo = EVEN,ODD
        do count = 1, bc_len_b(eo)
           i = bc_list_b(count, eo)
           do c2 = 1, NCOL
              do c1 = 1, NCOL
                 u(c1, c2, i, eo, mu1) = ZERO
              enddo
           enddo
!           do c2 = 1, NCOL
!              u(c2, c2, i, eo, mu1) = bound1(c2)
!           enddo
        enddo
!        do count = 1, bc_len_tt(eo)
!!write(STDERR,*)my_pe(),"ttttt"
!           i = bc_list_tt(count, eo)
!           do c2 = 1, NCOL
!              do c1 = 1, NCOL
!                 u(c1, c2, i, eo, mu1) = ZERO
!              enddo
!           enddo
!!           do c2 = 1, NCOL
!!              u(c2, c2, i, eo, mu1) = bound2(c2)
!!           enddo
!        enddo
     enddo
  enddo

!  do mu = 1, DIM
!     do eo = EVEN,ODD
!        do count = 1, bc_len_bb(eo)
!           i = bc_list_bb(count, eo)
!           do c2 = 1, NCOL
!              do c1 = 1, NCOL
!                 u(c1, c2, i, eo, mu) = ZERO
!              enddo
!           enddo
!        enddo
!     enddo
!  enddo

!STOP





  do mu = 1, DIM
     do e = EVEN, ODD
        o = EVEN + ODD - e
        do nu = mu + 1, DIM
           p = ZERO
           !$omp parallel do reduction(+: p) private(j1, j2, uuu)
           do i = 1, VOLH

              !  (j2,o) --<--   x        nu
              !    |            |        
              !    v            ^         ^
              !    |            |         |
              !  (i,e)  -->-- (j1,o)      x-->  mu


              j1 = nn(i, e, mu, FWD)
              j2 = nn(i, e, nu, FWD)
              
              uuu = 0
              call uuu_fwd(uuu, u(1, 1, j1, o, nu), &
                                u(1, 1, j2, o, mu), &
                                u(1, 1, i, e, nu))

              p = Re_Tr_uu(uuu, u(1, 1, i, e, mu))
if(p< 0.00001)write(STDERR,*)"zero u",my_pe(),i,&
"eo=",e,"munu=",mu,nu,u(1, 1, i, e, mu),u(1, 1, j1, o, nu), u(1, 1, j2, o, mu), u(1, 1, i, e, nu),j2

           enddo
        enddo
     enddo
  enddo
              
write(STDERR,*)sg(u)
STOP
end

!===============================================================================
