!===============================================================================
!
! clover_dsd.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine clover_dsd_w(eob, w_add, b, s, u)
  use typedef_clover
  use module_vol
  implicit none

  GAUGE_FIELD,intent(out) :: w_add
  integer,intent(in) :: eob    ! EVEN/ODD property of "b"
  CLOVER_FIELD_B,intent(in) :: b
  REAL,intent(in) :: s
  GAUGE_FIELD,intent(in) :: u
!!  GAUGE_FIELD :: w
  integer :: i, eo, mu, c1, c2

  COMPLEX, allocatable :: w(:, :, :, :, :) , t(:, :, :, :, :)
  integer :: ierr

  external :: sproj2_sigma_12   , sproj2_sigma_21
  external :: sproj2_sigma_31   , sproj2_sigma_13
  external :: sproj2_sigma_41_nr, sproj2_sigma_14_nr, sproj2_sigma_41_ch, sproj2_sigma_14_ch
  external :: sproj2_sigma_23   , sproj2_sigma_32
  external :: sproj2_sigma_24_nr, sproj2_sigma_24_ch, sproj2_sigma_42_nr, sproj2_sigma_42_ch
  external :: sproj2_sigma_43_nr, sproj2_sigma_34_nr, sproj2_sigma_43_ch, sproj2_sigma_34_ch

!!  CLOVER_FIELD_C2  :: t
!dir$ cache_align    t

  TIMING_START(timing_bin_clover_dsd_w)
  DEBUG2S("Start: clover_dsd_w")

  allocate(w(NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("clover_dsd: allocation failed w")
  endif
  allocate(t(NDIRAC, NDIRAC, NCOL, NCOL, volh), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("clover_dsd: allocation failed t")
  endif

  call clover_t_init(t, b(1, 1, eob))

#ifdef GAMMA_NOTATION_CHROMA
  call clovervv(w(1, 1, 1, EVEN, 1), t, sproj2_sigma_21   )
  call clovervv(w(1, 1, 1, ODD,  1), t, sproj2_sigma_31   )   ! <- 13
  call clovervv(w(1, 1, 1, EVEN, 2), t, sproj2_sigma_41_ch)
  call clovervv(w(1, 1, 1, ODD,  2), t, sproj2_sigma_32   )
  call clovervv(w(1, 1, 1, EVEN, 3), t, sproj2_sigma_42_ch)   ! <- 24 !s
  call clovervv(w(1, 1, 1, ODD,  3), t, sproj2_sigma_43_ch)           !s
#elif defined GAMMA_NOTATION_CHIRAL
  call clovervv(w(1, 1, 1, EVEN, 1), t, sproj2_sigma_12   )
  call clovervv(w(1, 1, 1, ODD,  1), t, sproj2_sigma_31   )  ! <- 13
  call clovervv(w(1, 1, 1, EVEN, 2), t, sproj2_sigma_14_ch)
  call clovervv(w(1, 1, 1, ODD,  2), t, sproj2_sigma_23   )
  call clovervv(w(1, 1, 1, EVEN, 3), t, sproj2_sigma_42_ch)  ! <- 24
  call clovervv(w(1, 1, 1, ODD,  3), t, sproj2_sigma_34_ch)
#elif defined GAMMA_NOTATION_DDHMC
  call clovervv(w(1, 1, 1, EVEN, 1), t, sproj2_sigma_12   )
  call clovervv(w(1, 1, 1, ODD,  1), t, sproj2_sigma_31   )  ! <- 13
  call clovervv(w(1, 1, 1, EVEN, 2), t, sproj2_sigma_41_ch)
  call clovervv(w(1, 1, 1, ODD,  2), t, sproj2_sigma_23   )
  call clovervv(w(1, 1, 1, EVEN, 3), t, sproj2_sigma_24_ch)  ! <- 24
  call clovervv(w(1, 1, 1, ODD,  3), t, sproj2_sigma_43_ch)
#else
  call clovervv(w(1, 1, 1, EVEN, 1), t, sproj2_sigma_12   )
  call clovervv(w(1, 1, 1, ODD,  1), t, sproj2_sigma_31   )  ! <- 13
  call clovervv(w(1, 1, 1, EVEN, 2), t, sproj2_sigma_41_nr)
  call clovervv(w(1, 1, 1, ODD,  2), t, sproj2_sigma_23   )
  call clovervv(w(1, 1, 1, EVEN, 3), t, sproj2_sigma_24_nr)  ! <- 42
  call clovervv(w(1, 1, 1, ODD,  3), t, sproj2_sigma_43_nr)
#endif

  w_add(1:NCOL, 1:NCOL, 1:volh, 0:1, 1:3) = &
  w_add(1:NCOL, 1:NCOL, 1:volh, 0:1, 1:3) + &
  s * w(1:NCOL, 1:NCOL, 1:volh, 0:1, 1:3)


  deallocate(w,  STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("clover_dsd: deallocation failed w")
  endif
  deallocate(t,  STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("clover_dsd: deallocation failed t")
  endif

  DEBUG2S("End: clover_dsd_w")
  TIMING_STOP(timing_bin_clover_dsd_w)
end

!-------------------------------------------------------------------------------
subroutine clovervv(w, t, sproj_sum)
  use module_vol
  implicit none
  COMPLEX, dimension(3, 3, volh_tot) :: w
  COMPLEX, dimension(4, 4, 3, 3, volh) :: t
  COMPLEX :: aa(4)
  external :: sproj_sum
  integer :: i, c1, c2

#ifdef PATHSCALE
  !$omp parallel do private(c1, c2, aa) shared(sproj_sum)
#else
  !$omp parallel do private(c1, c2, aa)
#endif
  do i = 1, volh
     do c2 = 1, 3
        do c1 = 1, 3
           call sproj_sum(aa, t(1, 1,c1, c2, i))
           w(c1, c2, i) = aa(1)+aa(2)+aa(3)+aa(4)
        enddo
     enddo
  enddo
end

!===============================================================================
