!===============================================================================
!
! sigma2.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine sproj2_sigma_12(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 1)
  out(2) =   in(2, 2)
  out(3) = - in(3, 3)
  out(4) =   in(4, 4)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_21(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) =   in(1, 1)
  out(2) = - in(2, 2)
  out(3) =   in(3, 3)
  out(4) = - in(4, 4)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_13(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = mi_times(in(1, 2))
  out(2) = pi_times(in(2, 1))
  out(3) = mi_times(in(3, 4))
  out(4) = pi_times(in(4, 3))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_31(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = pi_times(in(1, 2))
  out(2) = mi_times(in(2, 1))
  out(3) = pi_times(in(3, 4))
  out(4) = mi_times(in(4, 3))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_14_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 4)
  out(2) = - in(2, 3)
  out(3) = - in(3, 2)
  out(4) = - in(4, 1)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_41_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = in(1, 4)
  out(2) = in(2, 3)
  out(3) = in(3, 2)
  out(4) = in(4, 1)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_14_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) =   in(1, 2)
  out(2) =   in(2, 1)
  out(3) = - in(3, 4)
  out(4) = - in(4, 3)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_41_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 2)
  out(2) = - in(2, 1)
  out(3) =   in(3, 4)
  out(4) =   in(4, 3)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_23(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 2)
  out(2) = - in(2, 1)
  out(3) = - in(3, 4)
  out(4) = - in(4, 3)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_32(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) =  in(1, 2)
  out(2) =  in(2, 1)
  out(3) =  in(3, 4)
  out(4) =  in(4, 3)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_24_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = pi_times(in(1, 4))
  out(2) = mi_times(in(2, 3))
  out(3) = pi_times(in(3, 2))
  out(4) = mi_times(in(4, 1))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_42_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = mi_times(in(1, 4))
  out(2) = pi_times(in(2, 3))
  out(3) = mi_times(in(3, 2))
  out(4) = pi_times(in(4, 1))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_24_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = mi_times(in(1, 2))
  out(2) = pi_times(in(2, 1))
  out(3) = pi_times(in(3, 4))
  out(4) = mi_times(in(4, 3))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_42_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in
  ! statement function:
  COMPLEX :: pi_times, mi_times, c
  pi_times(c) = cmplx(-aimag(c), real(c), kind = RKIND)
  mi_times(c) = cmplx( aimag(c),-real(c), kind = RKIND)

  out(1) = pi_times(in(1, 2))
  out(2) = mi_times(in(2, 1))
  out(3) = mi_times(in(3, 4))
  out(4) = pi_times(in(4, 3))

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_34_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 3)
  out(2) =   in(2, 4)
  out(3) = - in(3, 1)
  out(4) =   in(4, 2)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_43_nr(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) =   in(1, 3)
  out(2) = - in(2, 4)
  out(3) =   in(3, 1)
  out(4) = - in(4, 2)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_43_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) = - in(1, 1)
  out(2) =   in(2, 2)
  out(3) =   in(3, 3)
  out(4) = - in(4, 4)

end
!-------------------------------------------------------------------------------
subroutine sproj2_sigma_34_ch(out,in)
  implicit none
  COMPLEX, dimension(4) :: out
  COMPLEX, dimension(4,4) :: in

  out(1) =   in(1, 1)
  out(2) = - in(2, 2)
  out(3) = - in(3, 3)
  out(4) =   in(4, 4)

end
!===============================================================================
