include(`defs.m4')
include(`spin_proj.m4')
include(`d100_defs.m4')
!===============================================================================
!
! D100.F90 - version for some gamma notations
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
#include "defs.h"
!-------------------------------------------------------------------------------
subroutine d(e, o, out, in, u)
  use module_d_g_reorder
  use module_vol
  use module_lattice
  use module_nn
  use module_d21
  use module_p_interface
  implicit none
  integer :: e, o
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: out, in
#else
  SPINCOL_FIELD :: out, in
#endif
  GAUGE_FIELD :: u

  TIMING_START(timing_bin_d)
  ALLOCATE_SC2_FIELD(a)  ! kept in 'module_d21'
  call d_projection(a, in)
  call xbound_sc2_field(a)
  call d_xf(out,a(1,1,1,mud(1),0),a(1,1,1, mud(1),1),u_reorder(1,1,1,1,e),nn(1,e,1,0),nn(1,e,1,1))
  call d_yf(out,a(1,1,1,mud(2),0),a(1,1,1, mud(2),1),u_reorder(1,1,1,2,e),nn(1,e,2,0),nn(1,e,2,1))
  call d_zf(out,a(1,1,1,mud(3),0),a(1,1,1, mud(3),1),u_reorder(1,1,1,3,e),nn(1,e,3,0),nn(1,e,3,1))
  call d_t( out,a(1,1,1,mud(4),0),a(1,1,1, mud(4),1),u_reorder(1,1,1,4,e),nn(1,e,4,0),nn(1,e,4,1))
  TIMING_STOP(timing_bin_d)
end

!-------------------------------------------------------------------------------
subroutine d_dag(e, o, out, in, u)
  use module_d_g_reorder
  use module_vol
  use module_lattice
  use module_nn
  use module_d21
  use module_p_interface
  implicit none
  integer :: e, o
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: out, in
#else
  SPINCOL_FIELD :: out, in
#endif
  GAUGE_FIELD :: u

  TIMING_START(timing_bin_d)
  ALLOCATE_SC2_FIELD(a)  ! kept in 'module_d21'
  call d_dag_projection(a, in)
  call xbound_sc2_field(a)
  call d_dag_xf(out,a(1,1,1,mud(1),0),a(1,1,1, mud(1),1),u_reorder(1,1,1,1,e),nn(1,e,1,0),nn(1,e,1,1))
  call d_dag_yf(out,a(1,1,1,mud(2),0),a(1,1,1, mud(2),1),u_reorder(1,1,1,2,e),nn(1,e,2,0),nn(1,e,2,1))
  call d_dag_zf(out,a(1,1,1,mud(3),0),a(1,1,1, mud(3),1),u_reorder(1,1,1,3,e),nn(1,e,3,0),nn(1,e,3,1))
  call d_dag_t( out,a(1,1,1,mud(4),0),a(1,1,1, mud(4),1),u_reorder(1,1,1,4,e),nn(1,e,4,0),nn(1,e,4,1))
  TIMING_STOP(timing_bin_d)
end

!-------------------------------------------------------------------------------
subroutine d_projection(out, in)
  use module_d_g_reorder
  use module_vol
  implicit none
  integer :: mu(4), i, col
  SC2_FIELD :: out
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: in
#else
  SPINCOL_FIELD :: in
#endif
  COMPLEX ci

  mu=mud
  ci = dcmplx(1,1)

# undef fwd
# undef bwd
# define fwd FWD
# define bwd BWD

  TIMING_START(timing_bin_sc2_projection)

#ifdef IBM
  call alignx(16, in(1,1,1) )
  call alignx(16, out(1,1,1,1,1))
#endif
  !$omp parallel do
  do i = 1, volh
_D_PROJECTION1
_D_PROJECTION2
_D_PROJECTION3
_D_PROJECTION4
  enddo
  TIMING_STOP(timing_bin_sc2_projection)

end

!-------------------------------------------------------------------------------
subroutine d_dag_projection(out, in)
  use module_d_g_reorder
  use module_vol
  implicit none
  integer :: mu(4), i, col
  SC2_FIELD :: out
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: in
#else
  SPINCOL_FIELD :: in
#endif
  COMPLEX ci

  mu=mud
  ci = dcmplx(1,1)

# undef fwd
# undef bwd
# define bwd FWD
# define fwd BWD

  TIMING_START(timing_bin_sc2_projection)
#ifdef IBM
  call alignx(16, in(1,1,1) )
  call alignx(16, out(1,1,1,1,1))
#endif
  !$omp parallel do
  do i = 1, volh
_D_PROJECTION1
_D_PROJECTION2
_D_PROJECTION3
_D_PROJECTION4
  enddo
  TIMING_STOP(timing_bin_sc2_projection)

end

!-------------------------------------------------------------------------------
subroutine d_block(out, in, u, eo, i1, i2)
  use module_d_g_reorder
  use module_vol
  use module_nn
  implicit none
  integer :: eo, i1, i2
  COMPLEX, dimension (4,3, volh_tot) :: out
  COMPLEX, dimension (2,3, volh_tot, DIM, 0:1) :: in
  GAUGE_FIELD :: u
  call d_x_block(out, in, u_reorder(1,1,1,1,eo),  nn(1,eo,1,FWD),nn(1,eo,1,BWD), i1,i2)
  call d_y_block(out, in, u_reorder(1,1,1,2,eo),  nn(1,eo,2,FWD),nn(1,eo,2,BWD), i1,i2)
  call d_z_block(out, in, u_reorder(1,1,1,3,eo),  nn(1,eo,3,FWD),nn(1,eo,3,BWD), i1,i2)
  call d_t_block(out, in, u_reorder(1,1,1,4,eo),  nn(1,eo,4,FWD),nn(1,eo,4,BWD), i1,i2)
end

!-------------------------------------------------------------------------------
subroutine d_dag_block(out, in, u, eo, i1, i2)
  use module_d_g_reorder
  use module_vol
  use module_nn
  implicit none
  integer :: eo, i1, i2
  COMPLEX, dimension (4,3, volh_tot) :: out
  COMPLEX, dimension (2,3, volh_tot, DIM, 0:1) :: in
  GAUGE_FIELD :: u
  call d_dag_x_block(out, in, u_reorder(1,1,1,1,eo),  nn(1,eo,1,FWD),nn(1,eo,1,BWD), i1,i2)
  call d_dag_y_block(out, in, u_reorder(1,1,1,2,eo),  nn(1,eo,2,FWD),nn(1,eo,2,BWD), i1,i2)
  call d_dag_z_block(out, in, u_reorder(1,1,1,3,eo),  nn(1,eo,3,FWD),nn(1,eo,3,BWD), i1,i2)
  call d_dag_t_block(out, in, u_reorder(1,1,1,4,eo),  nn(1,eo,4,FWD),nn(1,eo,4,BWD), i1,i2)
end

!-------------------------------------------------------------------------------
subroutine d_projection_block(out, in, i1, i2)
  use module_d_g_reorder
  use module_vol
  implicit none
  integer :: mu(4), i1, i2
  integer :: i
  SC2_FIELD :: out
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: in
#else
  SPINCOL_FIELD :: in
#endif
  COMPLEX ci

# undef fwd
# undef bwd
# define fwd FWD
# define bwd BWD

  mu=mud
  ci = dcmplx(1,1)

#ifdef IBM
  call alignx(16, in(1,1,1) )
  call alignx(16, out(1,1,1,1,1))
#endif
  do i = i1, i2
_D_PROJECTION1
_D_PROJECTION2
_D_PROJECTION3
_D_PROJECTION4
  enddo
end
!-------------------------------------------------------------------------------
subroutine d_dag_projection_block(out, in, i1, i2)
  use module_d_g_reorder
  use module_vol
  implicit none
  integer :: mu(4), i1, i2
  integer :: i
  SC2_FIELD :: out
#ifdef FLIPSC
  COMPLEX, dimension (NCOL, NDIRAC, volh_tot) :: in
#else
  SPINCOL_FIELD :: in
#endif
  COMPLEX ci

# undef fwd
# undef bwd
# define bwd FWD
# define fwd BWD

  mu=mud
  ci = dcmplx(1,1)
#ifdef IBM
  call alignx(16, in(1,1,1) )
  call alignx(16, out(1,1,1,1,1))
#endif
  do i = i1, i2
_D_PROJECTION1
_D_PROJECTION2
_D_PROJECTION3
_D_PROJECTION4
  enddo
end

!-------------------------------------------------------------------------------
subroutine d_xbound(a)

  use module_vol
  implicit none
  SC2_FIELD, intent(inout) :: a

  call xbound_sc2_field(a)

end

!===============================================================================
