!===============================================================================
!
! clover_action.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine clover_action_sum(sum)
  use module_switches
  use module_action
  use module_input
  implicit none
  REAL, intent(out) :: sum
  REAL :: tmp

  sum=0

  if (.not. switches%clover) return
  if (input%hmc_dsd == 0 )   return

#ifdef OMTDTD
  call clover_action_sum_eo(tmp, se)
  call clover_action_sum_eo(sum, so)
  sum = sum + tmp
#else
  call clover_action_sum_eo(sum, so)
#endif
end

!-------------------------------------------------------------------------------
subroutine clover_action_sum_eo(sum, eo)
  use module_field
  use module_action 
  use module_switches
  use module_vol
  implicit none
  REAL, intent(out) :: sum
  integer, intent(in) :: eo
  REAL, external    :: clover_action
  integer           :: i, cid
  REAL :: diff

  sum = ZERO
#ifdef GAMMAC
  diff = dble(volume) * log(64.0_8)
#else
  diff = ZERO
#endif

  do i = 1, size(action%logdet%cid)
     cid = action%logdet%cid(i)
     call init_clover(cid)
     sum = sum + ( &
                   clover_action(clover(cid)%b(1,1,eo))/dble(2) - diff &
                 ) * dble( clover(cid)%nf )
  enddo
end

!-------------------------------------------------------------------------------
subroutine dsd(p, step)
  use module_vol
  use module_switches
  use module_action
  implicit none
  GENERATOR_FIELD,  intent(inout) :: p
  REAL,             intent(in)    :: step

  if (step == ZERO .or. (.not. switches%clover)) return

  TIMING_START(timing_bin_clover_dsd)
  DEBUG2S("Start: dsd")

  call hmc_forces_old(p)
#ifdef OMTDTD
  call dsd_eo(p, step, se)
  call dsd_eo(p, step, so)
#else
  call dsd_eo(p, step, so)
#endif

  call nonactive(p) !! for ddhmc

  call hmc_forces_new(p, step, action%logdet%fid)

  DEBUG2S("End: dsd")
  TIMING_STOP(timing_bin_clover_dsd)

end

!-------------------------------------------------------------------------------
subroutine dsd_eo(p, step, eo)
  use module_field
  use module_action
  use module_vol
  use module_switches
  implicit none

  GENERATOR_FIELD,  intent(inout) :: p
  REAL,             intent(in)    :: step
  integer, intent(in) :: eo
  REAL :: s
  integer :: i, nf, cid

  COMPLEX, allocatable :: w(:, :, :, :, :)
  COMPLEX, allocatable :: ww(:, :, :, :, :)
!!  GAUGE_FIELD :: w, ww
  integer :: ierr

  allocate(w(NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("dsd: allocation failed w")
  endif
  allocate(ww(NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("dsd: allocation failed ww")
  endif

  w = ZERO
  do i = 1, size(action%logdet%cid)
     cid = action%logdet%cid(i)
     nf  = clover(cid)%nf
     call init_clover(cid)
     s = -step * clover(cid)%cswkappa * dble(nf) / EIGHT
     call clover_dsd_w(eo, w,  clover(cid)%b, s, gauge(3)%u(1,1,1,0,1))
  enddo

  call clover_d_w(ww, gauge(3)%u(1,1,1,0,1), w, eo)
  if (gauge(3)%sid /= 0) call dstout(ww, 1)
  call gen_m_force(p, ww)

  deallocate(w,  STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("dsd: deallocation failed w")
  endif
  deallocate(ww, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("dsd: deallocation failed ww")
  endif

end

!===============================================================================
