!===============================================================================
!
! dsf_mtmp.F90  --  p(j,x,mu) := p(j,x,mu) - step * D_{x,mu,j} S_{f}
!
! modified for all mass preconditioning terms, 28.04.2009, yn
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2009 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine dsf_mtmp(p, step, calc_sf, sf, fermiid)
  use module_action
  use module_field
  use module_hmc_forces
  use module_p_interface
  use module_switches
  use module_vol
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  integer,         intent(in)    :: fermiid, calc_sf
  REAL,            intent(in)    :: step
  REAL,            intent(out)   :: sf
!!  P_SPINCOL_FIELD, save          :: a, b
  COMPLEX, allocatable :: a(:,:,:)
  COMPLEX, allocatable :: b(:,:,:)
  REAL, external                 :: dotprod
  integer :: ierr

  sf = ZERO
  if (switches%quenched) return
  TIMING_START(timing_bin_dsf_mtmp)



  DEBUG2S("Start: dsf_mtmp")
  call init_quark(action%fermi(fermiid)%mid1)
  call init_quark(action%fermi(fermiid)%mid2)
!!  ALLOCATE_SC_FIELD(a)
!!  ALLOCATE_SC_FIELD(b)
  allocate(a(NDIRAC,NCOL,volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("dsf_mtmp: allocation failed a")
  endif
  allocate(b(NDIRAC,NCOL,volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("dsf_mtmp: allocation failed b")
  endif


  if (action%fermi(fermiid)%mid2 /=0 )  then
  TIMING_START(timing_bin_wdag_r8)
     call wdag_r8(b, mpf(fermiid)%sc(1,1,1), action%fermi(fermiid)%mid2)
  TIMING_STOP(timing_bin_wdag_r8)

  TIMING_START(timing_bin_cg_mre_mtmp)
#ifdef QPACE
     call solve_invmtdagmt(a, b, action%fermi(fermiid)%mid1, fermiid, calc_sf)
#else
     call cg_mre_mtmp(a, b, action%fermi(fermiid)%mid1, fermiid, calc_sf)
#endif
  TIMING_STOP(timing_bin_cg_mre_mtmp)

  else
  TIMING_START(timing_bin_cg_mre_mtmp)
#ifdef QPACE
     call solve_invmtdagmt(a, mpf(fermiid)%sc(1,1,1), action%fermi(fermiid)%mid1, fermiid, calc_sf)
#else
     call cg_mre_mtmp(a, mpf(fermiid)%sc(1,1,1), action%fermi(fermiid)%mid1, fermiid, calc_sf)
#endif
  TIMING_STOP(timing_bin_cg_mre_mtmp)
  endif

  TIMING_START(timing_bin_wmul_r8)
  call wmul_r8(b, a, action%fermi(fermiid)%mid1)
  TIMING_STOP(timing_bin_wmul_r8)

  if (calc_sf /= 0) sf = dotprod(b, b, SIZE_SC_FIELD)
  if (step == ZERO) return

  TIMING_START(timing_bin_dsf)
  call hmc_forces_old(p)
  if (action%fermi(fermiid)%mid2 ==0 ) then
     call dsf(p, step, a, b, action%fermi(fermiid)%mid1)
  else
     if (switches%mass_kappa) then
        call dsf(p, step, a, b                     ,action%fermi(fermiid)%mid1)
        call dsf(p,-step, a, mpf(fermiid)%sc(1,1,1),action%fermi(fermiid)%mid2)
     else
        call sc_axpy(b, mpf(fermiid)%sc(1,1,1), -ONE)
        call dsf(p, step, a, b, action%fermi(fermiid)%mid1)
     endif
  endif
  call hmc_forces_new(p, step, action%fermi(fermiid)%fid)
  TIMING_STOP(timing_bin_dsf)

  deallocate(a, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("dsf_mtmp: deallocation failed a")
  endif
  deallocate(b, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("dsf_mtmp: deallocation failed b")
  endif

  DEBUG2S("End: dsf_mtmp")
  TIMING_STOP(timing_bin_dsf_mtmp)
end

!===============================================================================
