!===============================================================================
!
! dsfr.F90  --  p(j,x,mu) := p(j,x,mu) - step * D_{x,mu,j} S_{fr}
!               for ration fractions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2006-2010 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine dsfr(frdo, p, step, calc_sf, fid)
  use module_field
  use module_action

  use module_input
  use module_switches
  use module_vol
  use module_sf
  use module_rhmc
  use module_hmc_forces
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  integer,         intent(in)    :: calc_sf, fid
  integer,         intent(inout) :: frdo(ratapp(0)%shift)
  REAL,            intent(in)    :: step(ratapp(0)%shift)
  COMPLEX, dimension(:,:,:,:,:),allocatable :: w, w1, w2
  integer :: i,nnn, mid , ierr, iteri
  REAL :: max_step

  if (switches%quenched) return
  DEBUG2S("Start: dsfr")
!-----------------------------------------------
  nnn=0
  do i=1,input%rhmc_roughness_level
       frdo(i)=1
  enddo
  max_step=ZERO
  do i=1,ratapp(0)%shift
     nnn = nnn  + frdo(i)
     max_step=max(max_step,abs(step(i)))
  enddo
  if ( nnn == ratapp(0)%shift ) return
  if (max_step == ZERO) return
!---------------------------
  allocate(w( NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr); w=ZERO
  allocate(w1(NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr); w1=ZERO
  allocate(w2(NCOL, NCOL, volh_tot, EVEN:ODD, DIM), STAT = ierr); w2=ZERO

  iteri=action%fermi(fid)%fid -3
  i=fid
  call hmc_forces_old(p)
  do while (action%fermi(i)%type == action%fermi(fid)%type .and. &
            action%fermi(i)%mid1 == action%fermi(fid)%mid1 )
     if (switches%measure_rhmc_forces)then
        call dsf_fraction_all(fid, frdo(1:ratapp(action%fermi(fid)%rid)%shift), &
             p, mpf(i)%sc, step(1:ratapp(action%fermi(fid)%rid)%shift), iteri)
     else
        call dsf_fraction_sep(fid, frdo(1:ratapp(action%fermi(fid)%rid)%shift), &
             p, mpf(i)%sc, step(1:ratapp(action%fermi(fid)%rid)%shift), iteri,w,w1,w2)
     endif
     i = i + 1
     if (i > size(action%fermi)) exit
  enddo

  if (.not. switches%measure_rhmc_forces)then
    call dsf_w2gen(p, w, w1, w2, action%fermi(fid)%mid1)
  endif
  call hmc_forces_new(p, max_step, action%fermi(fid)%fid )

  deallocate(w,  STAT = ierr)
  deallocate(w1, STAT = ierr)
  deallocate(w2, STAT = ierr)

  DEBUG2S("End: dsfr")
end

!-------------------------------------------------------------------------------
subroutine dsf_fraction_all(fid, frdo, p, in, step, np)
  use module_action
  use module_vol
  use module_rhmc
  use module_field
  implicit none

  GENERATOR_FIELD,intent(inout) :: p
  integer,        intent(in)    :: np, fid
  COMPLEX,        intent(in)    :: in(NDIRAC, NCOL, volh)
  REAL,           intent(in)    :: step(ratapp(action%fermi(fid)%rid)%shift)
  integer,        intent(in)    :: frdo(ratapp(action%fermi(fid)%rid)%shift)
  integer                       :: i, iters(ratapp(action%fermi(fid)%rid)%shift), mid, type

  REAL :: step_abar, factor, fe, bof(ratapp(action%fermi(fid)%rid)%shift)
  REAL, allocatable :: tmp(:), phi(:)
  integer :: ierr
  external :: wmul_r8, wtmul_all_dd, unprec_hmul
  external :: dsf, dsf_mtdagmt_dd, dsf_mtdagmt_hh

  mid = action%fermi(fid)%mid1
  type= action%fermi(fid)%type
  call init_quark(mid)

  factor=action%fermi(fid)%factor
  fe    =factor**ratapp(action%fermi(fid)%rid)%exponents(1) / factor
  bof   =ratapp(action%fermi(fid)%rid)%bbar/factor

  allocate( phi(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  allocate( tmp(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)

  call sc_copy(phi, in)
  call mcg_outer(type, phi, mid, ratapp(action%fermi(fid)%rid)%shift, frdo, bof, iters, np, CG_MD)
  do i=1,ratapp(action%fermi(fid)%rid)%shift
     if (frdo(i) /=1 ) then
        call rhmc_forces_old(p)
        step_abar=step(i)*ratapp(action%fermi(fid)%rid)%abar(i) * fe

        select case(type)
           case(EO_RATIONAL)
              call wmul_r8(tmp, sco(1,i)%sco(1), mid)
              call dsf(p, step_abar, sco(1,i)%sco(1), tmp, mid)
           case(DD_RATIONAL)
              call wtmul_all_dd(tmp, sco(1,i)%sco(1), mid)
              call dsf_mtdagmt_dd(p, step_abar, sco(1,i)%sco(1), tmp, mid)
           case(HH_RATIONAL)
              call unprec_hmul(tmp, sco(1,i)%sco(1), mid)
              call dsf_mtdagmt_hh(p, step_abar, sco(1,i)%sco(1), tmp, mid)
        end select

        call rhmc_forces_new(p, step(i)*fe, i, iters(i), type, mid)
     endif
  enddo

  deallocate(phi, STAT = ierr)
  deallocate(tmp, STAT = ierr)
end

!-------------------------------------------------------------------------------
subroutine dsf_fraction_sep(fid, frdo, p, in, step, np, w, w1, w2)
  use module_action
  use module_vol
  use module_rhmc
  use module_field
  implicit none

  GENERATOR_FIELD,intent(inout) :: p
  GAUGE_FIELD,    intent(inout) :: w, w1, w2
  integer,        intent(in)    :: np, fid
  COMPLEX,        intent(in)    :: in(NDIRAC, NCOL, volh)
  REAL,           intent(in)    :: step(ratapp(action%fermi(fid)%rid)%shift)
  integer,        intent(in)    :: frdo(ratapp(action%fermi(fid)%rid)%shift)
  integer                       :: i, iters(ratapp(action%fermi(fid)%rid)%shift), mid, type

  REAL :: step_abar, factor, fe, bof(ratapp(action%fermi(fid)%rid)%shift)
  REAL, allocatable :: tmp(:), phi(:)
  integer :: ierr
  external :: wmul_r8, wtmul_all_dd, unprec_hmul
  external :: dsf_sum, dsf_sum_dd, dsf_sum_hh

  mid = action%fermi(fid)%mid1
  type= action%fermi(fid)%type
  call init_quark(mid)

  factor=action%fermi(fid)%factor
  fe    =factor**ratapp(action%fermi(fid)%rid)%exponents(1) / factor
  bof   =ratapp(action%fermi(fid)%rid)%bbar/factor

  allocate( tmp(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  allocate( phi(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)

  call sc_copy(phi, in)
  call mcg_outer(type, phi, mid, ratapp(action%fermi(fid)%rid)%shift, frdo, bof, iters, np, CG_MD)
  do i=1,ratapp(action%fermi(fid)%rid)%shift
     if(frdo(i) /= 1 ) then
        step_abar=step(i)*ratapp(action%fermi(fid)%rid)%abar(i) * fe

        select case(type)
           case(EO_RATIONAL)
              call wmul_r8(tmp, sco(1,i)%sco(1), mid)
              call dsf_sum(w, w1, w2, step_abar, sco(1,i)%sco(1), tmp, mid)
           case(DD_RATIONAL)
              call wtmul_all_dd(tmp, sco(1,i)%sco(1), mid)
              call dsf_sum_dd(w, w1, w2, step_abar, sco(1,i)%sco(1), tmp, mid)
           case(HH_RATIONAL)
              call unprec_hmul(tmp, sco(1,i)%sco(1), mid)
              call dsf_sum_hh(w, w1, w2, step_abar, sco(1,i)%sco(1), tmp, mid)
        end select

     endif
  enddo

  deallocate(tmp, STAT = ierr)
  deallocate(phi, STAT = ierr)
end
!-----------------------------------------------------------------------------
subroutine rhmc_forces_old(p)
  use module_rhmc_forces
  use module_p_interface
  use module_switches
  use module_vol
  implicit none

  GENERATOR_FIELD, intent(in) :: p
  integer                     :: mu, eo, i

  if (.not. switches%measure_rhmc_forces) return

  if (.not. associated(p_old)) then
     call allocate_gen_field(p_old)
     f_count = ZERO
     f_avg = ZERO
     f_max = ZERO
     f_iter = ZERO
  endif

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do
        do i = 1, volh
           p_old(1, i, eo, mu) = p(1, i, eo, mu)
           p_old(2, i, eo, mu) = p(2, i, eo, mu)
           p_old(3, i, eo, mu) = p(3, i, eo, mu)
           p_old(4, i, eo, mu) = p(4, i, eo, mu)
           p_old(5, i, eo, mu) = p(5, i, eo, mu)
           p_old(6, i, eo, mu) = p(6, i, eo, mu)
           p_old(7, i, eo, mu) = p(7, i, eo, mu)
           p_old(8, i, eo, mu) = p(8, i, eo, mu)
        enddo
     enddo
  enddo

end

!-----------------------------------------------------------------------------
subroutine rhmc_forces_new(p, step, which, iter,  type, mid)

  use module_rhmc_forces
  use module_function_decl
  use module_switches
  use module_vol
  implicit none

  GENERATOR_FIELD, intent(in) :: p
  REAL, intent(in)            :: step
  integer, intent(in)         :: which, iter, type, mid

  integer                     :: mu, eo, i, iset
  REAL                        :: f, fa, fm, ffaa, ffmm

  if (.not. switches%measure_rhmc_forces) return

  ffaa = ZERO
  ffmm = ZERO
  do mu = 1, DIM
     do eo = EVEN, ODD
        fa = ZERO
        fm = ZERO
        !$omp parallel do reduction (+: fa) reduction(max: fm) private(f)
        do i = 1, volh
           f = sqrt( (p_old(1, i, eo, mu) - p(1, i, eo, mu))**2 &
                   + (p_old(2, i, eo, mu) - p(2, i, eo, mu))**2 &
                   + (p_old(3, i, eo, mu) - p(3, i, eo, mu))**2 &
                   + (p_old(4, i, eo, mu) - p(4, i, eo, mu))**2 &
                   + (p_old(5, i, eo, mu) - p(5, i, eo, mu))**2 &
                   + (p_old(6, i, eo, mu) - p(6, i, eo, mu))**2 &
                   + (p_old(7, i, eo, mu) - p(7, i, eo, mu))**2 &
                   + (p_old(8, i, eo, mu) - p(8, i, eo, mu))**2 )
           fa = fa + f
           fm = max(fm, f)
        enddo
        ffaa = ffaa + fa
        ffmm = max(ffmm, fm)
     enddo
  enddo

  ffaa = global_sum(ffaa) / (volume * DIM)
  ffmm = global_max(ffmm)

  iset=0
  do i=1,nset
     if (set(i,1)==type .and. set(i,2)==mid) iset=i
  enddo

  if (iset==0) then
     nset=nset+1
     set(nset,1)=type
     set(nset,2)=mid
     iset=nset
  endif

  f_count(which,iset) = f_count(which,iset) + ONE
    f_avg(which,iset) =   f_avg(which,iset) + ffaa / abs(step)
    f_max(which,iset) =   f_max(which,iset) + ffmm / abs(step)
   f_iter(which,iset) =  f_iter(which,iset) + real(iter, kind = RKIND)

end

!-----------------------------------------------------------------------------
subroutine rhmc_forces_write(unit)

  use module_rhmc
  use module_rhmc_forces
  use module_counter
  use module_function_decl
  use module_switches

  implicit none

  integer, intent(in)     :: unit
  integer, save           :: written = 0 
  integer                 :: i,j

  character(*), parameter :: key   = "%Frac"
  character(*), parameter :: fmt_h = "(1x, 2a,  3a6, a12, 3a20)"
  character(*), parameter :: fmt_b = "(1x, a6,  3i6, i12, 2g20.10, f20.3)"

  if (.not. switches%measure_rhmc_forces) return


  do j = 1, nset
  do i = 1, 40
     if (f_count(i,j) /= ZERO) then
           f_avg(i,j) =  f_avg(i,j) / f_count(i,j)
           f_max(i,j) =  f_max(i,j) / f_count(i,j)
          f_iter(i,j) = f_iter(i,j) / f_count(i,j)
     endif
  enddo
  enddo

  if (written == 0 .and. my_pe() == 0) then
     write(unit, fmt_h) "T", key, "traj", "type", "mid", "fraction", &
     "ave", "max", "iter"
     written = written + 1
  endif
     
  if (my_pe() == 0) then
     do j = 1, nset
     do i = 1, ratapp(0)%shift
        if (f_count(i,j) /= ZERO) then
           write(unit, fmt_b) key, counter%traj, set(j,1), set(j,2), i, &
           f_avg(i,j), f_max(i,j), f_iter(i,j)
        endif
     enddo
     enddo
  endif

  f_count = ZERO
  f_avg = ZERO
  f_max = ZERO
  f_iter = ZERO
end

!===============================================================================
