!===============================================================================
!
! w_mult.F90  -  W := M~ + rho (Hasenbusch improvement)
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2006-2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"
 
!-------------------------------------------------------------------------------
subroutine wdagw(out, in, id)  ! out = (W+ W) in
  use module_vol
  use module_p_interface
  implicit none
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(in)  :: in
  P_SPINCOL_FIELD, save :: tmp
  integer,intent(in) :: id

  TIMING_START(timing_bin_mtdagmt)
  ALLOCATE_SC_FIELD(tmp)
  call wmul(tmp(1,1,1), in,  id)
  call wdag(out, tmp(1,1,1), id)
  TIMING_STOP(timing_bin_mtdagmt)
end

!-------------------------------------------------------------------------------
subroutine wmul(out, in, id)  ! out = W in
  use module_vol
  use module_action
  implicit none
  integer,intent(in) :: id
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(in)  :: in

  call mtil(out, in, id)
  if(real(action%mtilde(id)%rho, kind = RKIND)  /= ZERO) then
     call sc_axpy(out, in, real(action%mtilde(id)%rho, kind = RKIND))
  endif
end

!-------------------------------------------------------------------------------
subroutine wdag(out, in, id)  ! out = W in
  use module_vol
  use module_action
  implicit none
  integer,intent(in) :: id
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(in)  :: in

  call mtil_dag(out, in, id)
  if(real(action%mtilde(id)%rho, kind = RKIND)  /= ZERO) then
     call sc_axpy(out, in, real(action%mtilde(id)%rho, kind = RKIND))
  endif
end

!-------------------------------------------------------------------------------
subroutine unprec_wdagw(oe, oo, ie, io, id)  ! out = (W+ W) in
  use module_vol
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  SPINCOL_FIELD :: tmp1, tmp2
  integer,intent(in) :: id

  TIMING_START(timing_bin_mtdagmt)

  call unprec_wmul(tmp1, tmp2, ie, io, id)
  call unprec_wdag(oe, oo, tmp1, tmp2, id)

  TIMING_STOP(timing_bin_mtdagmt)
end

!-------------------------------------------------------------------------------
subroutine unprec_wmul(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  SPINCOL_FIELD :: tmp
  integer,intent(in) :: id

  call d(se, so, oe, io, gauge(2)%u(1,1,1,0,1))
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_a(tmp, clover(action%mtilde(id)%cid)%a(1,1,se), ie, volh)
     call sc_xpby(oe, tmp, - action%mtilde(id)%kappa)
  else
     call sc_xpby(oe, ie, - action%mtilde(id)%kappa)
  endif


  call d(so, se, oo, ie, gauge(2)%u(1,1,1,0,1))
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_a(tmp, clover(action%mtilde(id)%cid)%a(1,1,so), io, volh)
     call sc_xpby(oo, tmp, - action%mtilde(id)%kappa)
  else
     call sc_xpby(oo, io, - action%mtilde(id)%kappa)
  endif


end

!-------------------------------------------------------------------------------
subroutine unprec_wdag(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  SPINCOL_FIELD :: tmp
  integer,intent(in) :: id

  call d_dag(se, so, oe, io, gauge(2)%u(1,1,1,0,1))
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_a(tmp, clover(action%mtilde(id)%cid)%a(1,1,se), ie, volh)
     call sc_xpby(oe, tmp, - action%mtilde(id)%kappa)
  else
     call sc_xpby(oe, ie, - action%mtilde(id)%kappa)
  endif
  call d_dag(so, se, oo, ie, gauge(2)%u(1,1,1,0,1))
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_a(tmp, clover(action%mtilde(id)%cid)%a(1,1,so), io, volh)
     call sc_xpby(oo, tmp, - action%mtilde(id)%kappa)
  else
     call sc_xpby(oo, io, - action%mtilde(id)%kappa)
  endif

end

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!
!  Tee in_e
!  Too in_o
!
!-------------------------------------------------------------------------------
subroutine unprec_cmul(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  integer,intent(in) :: id

  if (action%mtilde(id)%cid /=0)then
     call clover_mult_a(oe, clover(action%mtilde(id)%cid)%a(1,1,se), ie, volh)
     call clover_mult_a(oo, clover(action%mtilde(id)%cid)%a(1,1,so), io, volh)
  else
     oe = ie
     oo = io
  endif
end

!-------------------------------------------------------------------------------
!
! Tee^-1 Deo   
! or
! Too^-1 Doe   
!
!-------------------------------------------------------------------------------
subroutine mmul(out, in, eo, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(in)  :: in
  integer,intent(in) :: eo, id

  call d(1 - eo , eo, out, in, gauge(2)%u(1,1,1,0,1))
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_ao(clover(action%mtilde(id)%cid)%i(1,1, 1-eo ), out, volh)
  endif
  out = - action%mtilde(id)%kappa * out
end

!-------------------------------------------------------------------------------
!
! Deo^dagger Tee^-1
! or
! Doe^dagger Too^-1
!
!-------------------------------------------------------------------------------
subroutine mdag(out, in, eo, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(in)  :: in
  integer,intent(in) :: eo, id
  SPINCOL_FIELD :: tmp

  tmp = in
  if (action%mtilde(id)%cid /=0)then
     call clover_mult_ao(clover(action%mtilde(id)%cid)%i(1,1, eo ), tmp, volh)
  endif
  call d_dag(1 - eo , eo, out, tmp, gauge(2)%u(1,1,1,0,1))
  out = - action%mtilde(id)%kappa * out
end

!-------------------------------------------------------------------------------
!
! in_e + Meo in_o
! in_o + Moe in_e
!
!-------------------------------------------------------------------------------
subroutine unprec_mmul(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  integer,intent(in) :: id

  call mmul(oe, io, so , id)
  call sc_xpby(oe, ie, ONE)
  call mmul(oo, ie, se , id)
  call sc_xpby(oo, io, ONE)
end
!-------------------------------------------------------------------------------
!
! in_e + Moe^dagger in_o
! in_o + Meo^dagger in_e
!
!-------------------------------------------------------------------------------
subroutine unprec_mdag(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  integer,intent(in) :: id

  call mdag(oe, io, so , id)
  call sc_xpby(oe, ie, ONE)
  call mdag(oo, ie, se , id)
  call sc_xpby(oo, io, ONE)
end
!-------------------------------------------------------------------------------
subroutine unprec_mdagm(oe, oo, ie, io, id)
  use module_vol
  use module_action
  use module_field
  implicit none
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  SPINCOL_FIELD :: tmp1, tmp2
  integer,intent(in) :: id

  TIMING_START(timing_bin_mtdagmt)

  call unprec_mmul(tmp1, tmp2, ie, io, id)
  call unprec_mdag(oe, oo, tmp1, tmp2, id)

  TIMING_STOP(timing_bin_mtdagmt)
end
!===============================================================================
