!===============================================================================
!
! w_mult_block.F90  -  W := M~ + rho (Hasenbusch improvement)
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_m_mult_block

  REAL, dimension(:), pointer, save :: tmp1   ! sc2-field
  REAL, dimension(:), pointer, save :: tmp2   ! sc2-field

end

!-------------------------------------------------------------------------------
subroutine wdagw_block(out, in, id, extra_loops, r,  bk, paap)  ! out = (W+ W) in
  use module_vol
  use module_p_interface
  implicit none
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD, intent(inout)  :: in
  P_SPINCOL_FIELD, save :: tmp
  integer,intent(in) :: id
  logical,        intent(in)  :: extra_loops
  SPINCOL_FIELD,  intent(in)  :: r
  REAL,           intent(in)  :: bk
  DOUBLE,         intent(out) :: paap

  TIMING_START(timing_bin_mtdagmt)
  ALLOCATE_SC_FIELD(tmp)
  call wmul_block(tmp, in,  id, extra_loops, r,  bk, paap)
  call wdag_block(out, tmp, id, extra_loops, in, bk, paap)
  TIMING_STOP(timing_bin_mtdagmt)
end

!-------------------------------------------------------------------------------
subroutine wmul_block(out, in, id, extra_loops, p_or_r, bk, paap)

# undef DAGGER

# define M_MULT_D d
# define PLUS +
# define MINUS -

#ifdef OMTDTD
# include "m_mult_omtdtd_block.h90"
#else
# include "m_mult_block.h90"
#endif

end
!-------------------------------------------------------------------------------
subroutine wdag_block(out, in, id, extra_loops, p_or_r, bk, paap)
# define DAGGER

# undef M_MULT_D
# undef PLUS
# undef MINUS

# define M_MULT_D d_dag
# define PLUS -
# define MINUS +

#ifdef OMTDTD
# include "m_mult_omtdtd_block.h90"
#else
# include "m_mult_block.h90"
#endif
end
!===============================================================================
