!===============================================================================
!
! init_rhmc.F90 - program for RHMC
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2006-2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_rhmc(para)
  use module_field
  use module_p_interface

  use typedef_hmc
  use module_input
  use module_rhmc
  use module_switches
  use module_function_decl
  implicit none
  type(hmc_para),  intent(in)    :: para
  integer ::i, j, nth, max_frac, n_rid, iostat
  FILENAME:: list

  nth=0
  if (para%model=="F")then 
     nth=max(nth,input%hmc_mpf_rhmc_s*2)
     nth=max(nth,input%hmc_mpf_hh_rhmc_s*2)
     nth=max(nth,input%hmc_mpf_dd_rhmc_s*2)
  else
     nth=max(nth,input%hmc_mpf_rhmc_l)
     nth=max(nth,input%hmc_mpf_hh_rhmc_l)
     nth=max(nth,input%hmc_mpf_dd_rhmc_l)
  endif

  switches%rhmc=(nth>0)
  if ( .not. switches%rhmc ) return

  ratapp%max_rhmc =1
  ratapp%min_rhmc =1

!  write(0,*)"nth",nth

  select case (nth)
     case(2);call init_rhmc_2_e5_5(ratapp(0))
!!     case(4);call init_rhmc_4_e2_3()
!!     case(4);call init_rhmc_4_e4_4_v0()
!!     case(4);call init_rhmc_4_e4_4_v1()
!!     case(4);call init_rhmc_4_e5_5()
!!     case(4);call init_rhmc_4_e5_8()
!!     case(4);call init_rhmc_4_e8_8()
     case(4);call init_rhmc_4_e8_16(ratapp(0))
!!     case(4);call init_rhmc_4_e4_4()
!!     case(8);call init_rhmc_8_e6_4()
     case(8);call init_rhmc_8_e8_5(ratapp(0))
     case default
        call die("wrong rhmc")
  end select

  max_frac = max(ratapp(0)%shift_exact, ratapp(0)%shift)

  if (.not. associated(sco)) then
     allocate(sco(2, max_frac))
     do i = 1, 2
        do j = 1, max_frac
           nullify(sco(i, j)%sco)
           call allocate_sc_overindexed(sco(i, j)%sco)
        enddo
     enddo
  endif

  if (input%tuning_approx_range /=0) then
     list = input%tuning_approx_range_list
     if (list /= " ") then
        open(ULIST, file = list, action = "read", status = "old")
        iostat = 0
        n_rid  = 0
        do while (iostat == 0)
           read(ULIST, *, iostat = iostat) i
        enddo
        n_rid = i


        rewind(ULIST)

        iostat = 0
        do j = 1, n_rid
           read(ULIST, *) i, ratapp(j)%shift, ratapp(j)%shift_exact, ratapp(j)%margin_factor
           if (i/=j) call die("list is not correct format, skipped rid")
        enddo
        close(ULIST)
     else
        call die("init_rhmc(): list is missing")
     endif
  endif

  ratapp(0)%time_switch=1
!!  do i=1,min(ratapp(0)%shift,input%rhmc_fraction_u1)
!!     ratapp(0)%time_switch(i)=1
!!  enddo
!!  do i=min(ratapp(0)%shift,input%rhmc_fraction_u1)+1,min(ratapp(0)%shift,input%rhmc_fraction_u1+input%rhmc_fraction_u2)
!!     ratapp(0)%time_switch(i)=2
!!  enddo

end

!-------------------------------------------------------------------------------
subroutine gene_new_coeff()
  use module_action
  use module_rhmc
  use module_input
  implicit none
  REAL,allocatable :: dr(:,:),dp(:,:)
  REAL :: min_new, max_new
  REAL :: min_ev, max_ev, factor
  integer :: fid, mid, rid, ftype, min_i, max_i, frid, deg(2), pm, num, check_nth
  logical :: id_done(n_mtilde_id,8)
  logical, external :: type_rational

  id_done = .false.
  min_new =  99999
  min_ev  =  99999
  max_new = -99999
  max_ev  = -99999

  rid   = 0

!! check when tuning_approx_range,
!! "non zero nth" of all of ftype must be same
if (input%tuning_approx_range==0) then

  check_nth=0
  do fid = 1, size(action%fermi)
     ftype = action%fermi(fid)%type
     mid   = action%fermi(fid)%mid1
     if ( type_rational(ftype) ) then
        if (check_nth==0) then
           check_nth=action%fermi(fid)%nth
        else
           if (check_nth/=action%fermi(fid)%nth) then
              write(0,*) "ERROR:with tuning_approx_range=0, different nth", &
              check_nth,action%fermi(fid)%nth
              call die("with tuning_approx_range=0: different nth")
           endif
        endif
     endif
  enddo
endif

  do fid = 1, size(action%fermi)
     ftype = action%fermi(fid)%type
     mid   = action%fermi(fid)%mid1
     if ( type_rational(ftype) ) then
        if (.not. id_done(mid,ftype)) then
           call minmax_id(ftype, mid, min_ev, max_ev, min_i, max_i)
           id_done(mid,ftype) = .true.
           if (input%tuning_approx_range/=0)rid = rid + 1
        endif
        if (max_ev/min_ev > ratapp(rid)%max_rhmc/ratapp(rid)%min_rhmc) then
           ratapp(rid)%max_rhmc      = (max_ev / min_ev) * ratapp(rid)%margin_factor
           ratapp(rid)%min_rhmc      = 1
           deg(1) = ratapp(rid)%shift
           deg(2) = ratapp(rid)%shift_exact

           num=action%fermi(fid)%nth
           do pm = 1, 2
           allocate(dr(0:deg(pm), pm))
           allocate(dp(1:deg(pm), pm))
#ifdef FMLIB
           call remez(ratapp(rid)%min_rhmc,ratapp(rid)%max_rhmc, &
                      1, num*pm, deg(pm), 60, dr, dp, pm)
#else
           call die("remez method needs FMLIB")
#endif

           call storeremez(rid, deg(pm), pm, num*pm, dr, dp)
           deallocate(dr)
           deallocate(dp)
           enddo
        endif

        action%fermi(fid)%rid = rid
        frid = action%fermi(fid)%rid

        !! re-scale as to be
        !!                    log(factor*min_ev)-log(min_rhmc) =
        !!                    log(max_rhmc)-log(factor*max_ev)
        factor= sqrt(ratapp(frid)%max_rhmc*ratapp(frid)%min_rhmc/max_ev/min_ev)
        action%fermi(fid)%factor = factor
     endif
  enddo

  DEBUG2S("End: need_new_coeff")
  !!         if (          factor*min_ev < ratapp(rid)%min_rhmc) then
  !!            write(0,*)"factor*min_ev < min_rhmc, f*min_ev, min_r",factor*min_ev,ratapp(rid)%min_rhmc
  !!             call die("factor*min_ev < min_rhmc")
  !!         endif
  !!         if (          factor*max_ev > ratapp(rid)%max_rhmc) then
  !!            write(0,*)"factor*max_ev > max_rhmc, f*max_ev, max_r",factor*max_ev,ratapp(rid)%max_rhmc
  !!            call die("factor*max_ev > max_rhmc")
  !!         endif
  !!      if (max_ev/min_ev > ratapp(frid)%max_rhmc/ratapp(frid)%min_rhmc) then
  !!      endif
end

!-------------------------------------------------------------------------------
subroutine storeremez(rid, deg, pm, num, dr, dp)
  use module_rhmc
  implicit none
  integer :: rid,deg,pm,num
  real(8) :: dr(0:deg,pm),dp(deg,pm)

  if (pm == 1) then
     ratapp(rid)%shift = deg
     ratapp(rid)%exponents(1) = ONE/dble(num)

     if (associated(ratapp(rid)%time_switch)) deallocate(ratapp(rid)%time_switch)
     allocate(ratapp(rid)%time_switch(1:deg))

     if (associated(ratapp(rid)%abar)) deallocate(ratapp(rid)%abar)
     if (associated(ratapp(rid)%bbar)) deallocate(ratapp(rid)%bbar)
     allocate(ratapp(rid)%abar(0:deg))
     allocate(ratapp(rid)%bbar(1:deg))

     ratapp(rid)%abar(0:deg) = dr(0:deg,1)
     ratapp(rid)%bbar(1:deg) = dp(1:deg,1)

  else
     ratapp(rid)%shift_exact = deg
     ratapp(rid)%exponents(2) = ONE/dble(num)

     if (associated(ratapp(rid)%abar_exact_p)) deallocate(ratapp(rid)%abar_exact_p)
     if (associated(ratapp(rid)%bbar_exact_p)) deallocate(ratapp(rid)%bbar_exact_p)
     allocate(ratapp(rid)%abar_exact_p(0:deg))
     allocate(ratapp(rid)%bbar_exact_p(1:deg))
     if (associated(ratapp(rid)%abar_exact_m)) deallocate(ratapp(rid)%abar_exact_m)
     if (associated(ratapp(rid)%bbar_exact_m)) deallocate(ratapp(rid)%bbar_exact_m)
     allocate(ratapp(rid)%abar_exact_m(0:deg))
     allocate(ratapp(rid)%bbar_exact_m(1:deg))

     ratapp(rid)%abar_exact_m(0:deg) = dr(0:deg,1)
     ratapp(rid)%bbar_exact_m(1:deg) = dp(1:deg,1)
     ratapp(rid)%abar_exact_p(0:deg) = dr(0:deg,2)
     ratapp(rid)%bbar_exact_p(1:deg) = dp(1:deg,2)

  endif

end
!===============================================================================
