!===============================================================================
!
! program for RHMC
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_rhmc_4_e2_3(rin)

  use module_rhmc
  implicit none
  type(rational_approx), intent(out) :: rin

  rin%shift = 20
  rin%shift_exact = 40
  rin%min_rhmc = 0.01_8
  rin%max_rhmc = 3.0_8
  rin%exponents(1) = 0.25_8
  rin%exponents(2) = 0.125_8

  if (.not. associated(rin%abar)) then

  allocate(rin%time_switch( 1:rin%shift))
  allocate(rin%abar(        0:rin%shift)) 
  allocate(rin%bbar(        1:rin%shift))
  allocate(rin%abar_exact_p(0:rin%shift_exact))
  allocate(rin%bbar_exact_p(1:rin%shift_exact))
  allocate(rin%abar_exact_m(0:rin%shift_exact))
  allocate(rin%bbar_exact_m(1:rin%shift_exact))

  endif

  call init_rhmc_m4_20_e2_3(rin%shift      ,rin%abar,        rin%bbar)
  call init_rhmc_m8_40_e2_3(rin%shift_exact,rin%abar_exact_m,rin%bbar_exact_m)
  call init_rhmc_8_40_e2_3( rin%shift_exact,rin%abar_exact_p,rin%bbar_exact_p)

end
!======================================================================== 4root
subroutine init_rhmc_m4_20_e2_3(n,alpha,beta)
integer::n
REAL,intent(out)::alpha(0:n)
REAL,intent(out)::beta(1:n)
!!./remez.x 1 4 20 20 0.01 3  100
!!Converged at 899 iterations, error = 5.136776d-21
!!Approximation to f(x) = x^(-1/4)
alpha(0) = 2.0395586656184123d-01
alpha(1) = 1.0616169296361420d-03; beta(1) = 1.7554233252991956d-04
alpha(2) = 1.8199558393076232d-03; beta(2) = 1.1819110884261391d-03
alpha(3) = 2.5604626355205055d-03; beta(3) = 3.2610155973352511d-03
alpha(4) = 3.4618920672159024d-03; beta(4) = 6.7763046372399660d-03
alpha(5) = 4.6469417258854055d-03; beta(5) = 1.2343410584705609d-02
alpha(6) = 6.2508260093438687d-03; beta(6) = 2.0941884517032393d-02
alpha(7) = 8.4481007801424923d-03; beta(7) = 3.4095832949177125d-02
alpha(8) = 1.1478972246924745d-02; beta(8) = 5.4163127592831165d-02
alpha(9) = 1.5686380294737699d-02; beta(9) = 8.4802198037480972d-02
alpha(10) = 2.1576245610835886d-02; beta(10) = 1.3174192757848904d-01
alpha(11) = 2.9923875760713141d-02; beta(11) = 2.0409536190234331d-01
alpha(12) = 4.1975836642918721d-02; beta(12) = 3.1670931842606848d-01
alpha(13) = 5.9861915868044666d-02; beta(13) = 4.9463293361373772d-01
alpha(14) = 8.7506889447362221d-02; beta(14) = 7.8230067491116317d-01
alpha(15) = 1.3285427250508156d-01; beta(15) = 1.2643154267152963d+00
alpha(16) = 2.1400907091480129d-01; beta(16) = 2.1185891625728952d+00
alpha(17) = 3.7935586807136867d-01; beta(17) = 3.7760663038648601d+00
alpha(18) = 7.9220852746359494d-01; beta(18) = 7.5261014438793339d+00
alpha(19) = 2.2699644122937053d+00; beta(19) = 1.8862270647721509d+01
alpha(20) = 1.4680710560420376d+01; beta(20) = 8.8477489789401588d+01
end
!======================================================================== 8root
subroutine init_rhmc_8_40_e2_3(n,alpha,beta)
integer::n
REAL,intent(out)::alpha(0:n)
REAL,intent(out)::beta(1:n)
!!./remez.x 1 8 40 40 0.01 3  100
!!Converged at 3655 iterations, error = 1.638322d-41
!!Approximation to f(x) = x^(1/8)
alpha(0) = 2.6123257916249929d+00
alpha(1) = -6.6661265804105085d-06; beta(1) = 7.5004903713937697d-05
alpha(2) = -1.8455761021699046d-05; beta(2) = 3.6718414343745647d-04
alpha(3) = -3.3098298668191518d-05; beta(3) = 8.9550070215212814d-04
alpha(4) = -5.0890186035981347d-05; beta(4) = 1.6833874800451260d-03
alpha(5) = -7.2584205338750460d-05; beta(5) = 2.7656970905211424d-03
alpha(6) = -9.9258052147894785d-05; beta(6) = 4.1903389536960762d-03
alpha(7) = -1.3232508215366213d-04; beta(7) = 6.0204533143655781d-03
alpha(8) = -1.7359513829130345d-04; beta(8) = 8.3372832355026964d-03
alpha(9) = -2.2537035156711379d-04; beta(9) = 1.1243892475696881d-02
alpha(10) = -2.9057641591059488d-04; beta(10) = 1.4869919917818757d-02
alpha(11) = -3.7293618025544277d-04; beta(11) = 1.9377620411211542d-02
alpha(12) = -4.7719714484674084d-04; beta(12) = 2.4969520027983347d-02
alpha(13) = -6.0942971526579525d-04; beta(13) = 3.1898118024618630d-02
alpha(14) = -7.7742000999526148d-04; beta(14) = 4.0478209183599755d-02
alpha(15) = -9.9119081378342441d-04; beta(15) = 5.1102594952793574d-02
alpha(16) = -1.2636985576505620d-03; beta(16) = 6.4262224358413125d-02
alpha(17) = -1.6117755130705604d-03; beta(17) = 8.0572193337452330d-02
alpha(18) = -2.0574187657191219d-03; beta(18) = 1.0080559166230765d-01
alpha(19) = -2.6295776424945768d-03; beta(19) = 1.2593801083692704d-01
alpha(20) = -3.3666702723111749d-03; beta(20) = 1.5720675949729876d-01
alpha(21) = -4.3201869917233269d-03; beta(21) = 1.9619071154865111d-01
alpha(22) = -5.5599468594727999d-03; beta(22) = 2.4491962958931132d-01
alpha(23) = -7.1819237631712040d-03; beta(23) = 3.0602642956945769d-01
alpha(24) = -9.3201614600208968d-03; beta(24) = 3.8296334332007065d-01
alpha(25) = -1.2165363454074195d-02; beta(25) = 4.8031536899503141d-01
alpha(26) = -1.5994689064301029d-02; beta(26) = 6.0426559840582239d-01
alpha(27) = -2.1220959467402208d-02; beta(27) = 7.6330425748811626d-01
alpha(28) = -2.8476684114257800d-02; beta(28) = 9.6934100221459107d-01
alpha(29) = -3.8763098501924373d-02; beta(29) = 1.2395079406809963d+00
alpha(30) = -5.3726296066653800d-02; beta(30) = 1.5991935565595119d+00
alpha(31) = -7.6195519294770711d-02; beta(31) = 2.0873733052154830d+00
alpha(32) = -1.1129764215691378d-01; beta(32) = 2.7664640233352253d+00
alpha(33) = -1.6893847025999353d-01; beta(33) = 3.7416872229554681d+00
alpha(34) = -2.6984472173312213d-01; beta(34) = 5.2020740348085308d+00
alpha(35) = -4.6204416335645992d-01; beta(35) = 7.5158959519926745d+00
alpha(36) = -8.7304631199469773d-01; beta(36) = 1.1481909195252499d+01
alpha(37) = -1.9129418902693256d+00; beta(37) = 1.9112746895948721d+01
alpha(38) = -5.3518084064797238d+00; beta(38) = 3.6771448862894459d+01
alpha(39) = -2.4093139261064344d+01; beta(39) = 9.4305838669014065d+01
alpha(40) = -4.1403990534625950d+02; beta(40) = 5.5239376625635600d+02
end
!========================================================================
subroutine init_rhmc_m8_40_e2_3(n,alpha,beta)
integer::n
REAL,intent(out)::alpha(0:n)
REAL,intent(out)::beta(1:n)
!!./remez.x 1 8 40 40 0.01 3  100
!!Converged at 3655 iterations, error = 1.638322d-41
!!Approximation to f(x) = x^(-1/8)
alpha(0) = 3.8280064577165612d-01
alpha(1) = 6.3099477149094988d-05; beta(1) = 5.4309084990791766d-05
alpha(2) = 1.2597875889923537d-04; beta(2) = 3.1811391981032304d-04
alpha(3) = 1.8406031920197605d-04; beta(3) = 8.1585036564258339d-04
alpha(4) = 2.4352113443747630d-04; beta(4) = 1.5696330916389117d-03
alpha(5) = 3.0795705947575812d-04; beta(5) = 2.6128058922817734d-03
alpha(6) = 3.8036720732295785d-04; beta(6) = 3.9915400893816471d-03
alpha(7) = 4.6370520203408534d-04; beta(7) = 5.7669306125329280d-03
alpha(8) = 5.6114447776737793d-04; beta(8) = 8.0177733232078465d-03
alpha(9) = 6.7626515114588043d-04; beta(9) = 1.0844167770463993d-02
alpha(10) = 8.1322417561896629d-04; beta(10) = 1.4372129759943947d-02
alpha(11) = 9.7693560126074579d-04; beta(11) = 1.8759455274783426d-02
alpha(12) = 1.1732780774153105d-03; beta(12) = 2.4203152731331227d-02
alpha(13) = 1.4093450967399559d-03; beta(13) = 3.0948861062784848d-02
alpha(14) = 1.6937556004138637d-03; beta(14) = 3.9302807112230824d-02
alpha(15) = 2.0370475171225750d-03; beta(15) = 4.9647042755944087d-02
alpha(16) = 2.4521849337641437d-03; beta(16) = 6.2458963290659003d-02
alpha(17) = 2.9552220737248018d-03; beta(17) = 7.8336479256519320d-02
alpha(18) = 3.5661862228112775d-03; beta(18) = 9.8030748658559930d-02
alpha(19) = 4.3102707445706115d-03; beta(19) = 1.2248916124160776d-01
alpha(20) = 5.2194742358925243d-03; beta(20) = 1.5291243791916542d-01
alpha(21) = 6.3348925172698753d-03; beta(21) = 1.9083148902713362d-01
alpha(22) = 7.7099833040728130d-03; beta(22) = 2.3821243324897365d-01
alpha(23) = 9.4153083705876845d-03; beta(23) = 2.9760253876092607d-01
alpha(24) = 1.1545567348617683d-02; beta(24) = 3.7233689139321369d-01
alpha(25) = 1.4230268010176825d-02; beta(25) = 4.6683724846932473d-01
alpha(26) = 1.7650314921101123d-02; beta(26) = 5.8705433701972931d-01
alpha(27) = 2.2064507181551091d-02; beta(27) = 7.4113950703517961d-01
alpha(28) = 2.7853168268516273d-02; beta(28) = 9.4049435696633632d-01
alpha(29) = 3.5592504273584194d-02; beta(29) = 1.2014648245692747d+00
alpha(30) = 4.6186465428371293d-02; beta(30) = 1.5481777103365355d+00
alpha(31) = 6.1111676959902390d-02; beta(31) = 2.0174957340591146d+00
alpha(32) = 8.2898087710172735d-02; beta(32) = 2.6681151625065360d+00
alpha(33) = 1.1613671690286191d-01; beta(33) = 3.5982944506731940d+00
alpha(34) = 1.6977153843646228d-01; beta(34) = 4.9830134764796092d+00
alpha(35) = 2.6287454732166360d-01; beta(35) = 7.1593253747500754d+00
alpha(36) = 4.4127975026178917d-01; beta(36) = 1.0847174877834174d+01
alpha(37) = 8.3511628125207071d-01; beta(37) = 1.7821208934734265d+01
alpha(38) = 1.9193532262074948d+00; beta(38) = 3.3500811253304391d+01
alpha(39) = 6.3656861021018365d+00; beta(39) = 8.1702874528158887d+01
alpha(40) = 5.5103011948229089d+01; beta(40) = 3.9997384856885418d+02
end
!========================================================================
