!===============================================================================
!
! cg_dd.F90 
!
! solves "matrix_mult * x = b" and returns number of iterations
!
! for DD, 09.2009 ,yn
!-------------------------------------------------------------------------------
!
! Copyright (C) 2009 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine cg_unprec_mdagm_dd(x, b, id, iter, res, np, calc_sf) 
  use module_cg
  use module_dd
  implicit none

  REAL,dimension(24*ld%lat%volh,0:ld%lat%num_ip-1,0:1),intent(inout) :: x
  REAL,dimension(24*ld%lat%volh,0:ld%lat%num_ip-1,0:1),intent(in)    :: b

  integer,              intent(out)   :: iter
  integer,              intent(in)    :: id, np, calc_sf
  DOUBLE,               intent(in)    :: res

  REAL,dimension(24*ld%lat%volh,0:ld%lat%num_ip-1,0:1) :: r,p,aap
  REAL          :: ak, bk
  DOUBLE        :: rtr, rtrold, paap, bb
  integer       :: i, niter
  REAL, external :: ddsc_norm2_eo
  COMPLEX, external :: ddsc_cdotc_eo

  character(72) :: msg

!
! needed for checking solution
!
  DOUBLE        :: norm(3)
  integer, save :: headline_check_written = 0
  character(*), parameter :: key_chke = "%cgChkEven"
  character(*), parameter :: key_chko = "%cgChkOdd "
  character(*), parameter :: fmt_chk_h = "(1x, 2a,  a6, 4a20)"
  character(*), parameter :: fmt_chk_b = "(1x, a11, i6, 3e20.10, i20)"

!-------------------------------------------------------------------------------
  TIMING_START(timing_bin_cg)
  DEBUG2S("Start: cg_hh")


  bb = ddsc_norm2_eo(b, 0);  call global_sum_vec(1, bb)
  call unprec_mdagm_dd(r, x, id)
  call ddsc_sub2_eo(r,b,0) ! r = b - r
  call ddsc_copy_eo(p,r,0) ! p = r

  rtrold = ddsc_norm2_eo(r,0);  call global_sum_vec(1, rtrold)
  if (stop_condition(rtrold, res, bb)) goto 9999

  do niter = 1, cg_para%maxiter
     call unprec_mdagm_dd(aap, p, id)
     paap = dble( ddsc_cdotc_eo(p,aap,0)  );   call global_sum_vec(1, paap)

     if (paap /= 0) ak = rtrold / paap 

     call ddsc_axpy_eo(x,   p,  ak, 0)
     call ddsc_axpy_eo(r, aap, -ak, 0)

     rtr = ddsc_norm2_eo(r,0);  call global_sum_vec(1, rtr)

     if (stop_condition(rtr, res, bb)) goto 9999
     if (rtrold /= 0) bk = rtr / rtrold
     rtrold = rtr

     call ddsc_xpby_eo(p, r, bk, 0)! p = r + bk * p
  enddo

  niter = niter - 1

  if (cg_para%log /= 2) then
     write(msg, *) "cg_dd(): no convergence; rtr = ", rtr 
     call die(msg)
  endif

9999 continue
  call update_cg_stat(cg_stat, niter, 0)
  cg_iterations_total = cg_stat%niter_tot_run

  iter = niter

  TIMING_STOP(timing_bin_cg)
  DEBUG2S("End: cg_hh")

end

!===============================================================================
