!===============================================================================
!
! coefficient_sig.F90 - coefficients for gauge action
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2006-2009 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine coefficient_sig(para)
  use module_input
  use typedef_hmc
  use module_counter
  use module_action
  use module_switches
  use module_function_decl
  implicit none
  type(hmc_para), intent(inout) :: para
  integer :: counter_traj, icount
  REAL, external :: calculate_plaq_ave, csw1_alpha
  REAL :: csw

  if (gauge_action == "TREETAD" .or. switches%tadcsw) then
     icount = 0
     csw = para%csw
     counter_traj = counter%traj
     if (switches%fixtad) then
        if (switches%fixtad_u04) then
           read(input%u04(1), *) plaq_ave
        else
           plaq_ave = ONE
        endif
     else
        if (counter%j_traj == 0) counter_traj = counter_traj+1
        plaq_ave = calculate_plaq_ave(counter_traj,icount)
     endif
  endif

  if (counter%j_traj == 0 .or. &
     (gauge_action=="TREETAD" .and. (.not. switches%fixtad)) &
     ) then
     select case (gauge_action)
        case ("WILSON")  ; call coefficient_wilson(para)
        case ("TREE")    ; call coefficient_tree(para)
        case ("IWASAKI") ; call coefficient_iwasaki(para)
        case ("TREETAD") ; call coefficient_treetad(para)
        case default
           call die("coefficient_sig: " // gauge_action // ": unknown action")
     end select
     call coefficient_beta0123(para%beta)
  endif

  if (switches%tadcsw) then
     if (counter%j_traj == 0 .or. (.not. switches%fixtad)) then
        csw = ONE + g2 * csw1_alpha(para%alpha)
        csw = csw/(plaq_ave)**(0.75_8)
        call coefficient_csw_kappa(csw, para)
     endif
  endif

  if (gauge_action == "TREETAD" .or. switches%tadcsw) then
     if (counter%j_traj /= 0) then
        if(my_pe()==0)write(UREC,200)"%Tad", &
                      counter_traj, &
                      counter_traj - icount, &
                      counter_traj - 1, icount, &
                      action%plaq%beta, &
                      action%impg%beta1, &
                      action%impg%beta2, &
                      action%impg%beta3, &
                      csw,plaq_ave
     endif
  endif
  200 format (2x, a4, 1x, 4i6, 6g15.8)

end

!-------------------------------------------------------------------------------
REAL function calculate_plaq_ave(n,icount)

  use module_action
  use module_function_decl
  implicit none

  integer, intent(in)  :: n
  integer, intent(out) :: icount
  integer :: i

  calculate_plaq_ave = ZERO
  icount = 0

  do i = n-last_n_ave_plaq, n-1
     if (i >= plaq_start) then
        calculate_plaq_ave = calculate_plaq_ave + plaq_hist(i)
        icount = icount + 1
     endif
  enddo
  if (icount == 0) then
     calculate_plaq_ave = ONE
  else
     calculate_plaq_ave = calculate_plaq_ave / dble(icount)
  endif
  if (icount==0 .or. icount /= last_n_ave_plaq) then
     if (my_pe() == 0) write(UREC,*) "WARNING: TAD_sig  :", &
                   "icount=",icount,"n_ave_plaq=",last_n_ave_plaq
  endif

end

!-------------------------------------------------------------------------------
subroutine coefficient_beta0123(para_beta)
  use module_action
  implicit none
  REAL, intent(in) :: para_beta

  action%plaq%beta =para_beta*c0
  action%impg%beta1=para_beta*c1
  action%impg%beta2=para_beta*c2
  action%impg%beta3=para_beta*c3
end

!-------------------------------------------------------------------------------
subroutine coefficient_csw_kappa(csw, para)
  use module_action
  use typedef_hmc
  implicit none
  type(hmc_para), intent(inout) :: para
  REAL, intent(in) :: csw

  if (csw == para%csw ) return

  para%csw               = csw
  para%csw_kappa         = csw * para%kappa
  para%csw_kappa_strange = csw * para%kappa_strange

  n_clover_id = 0
  n_mtilde_id = 0
  n_dslash_id = 0
  n_fermi_id = 0
  call init_clover_field(para)
  call init_fermi_field(para)

end

!-------------------------------------------------------------------------------
subroutine coefficient_wilson(para)
  use typedef_hmc
  use module_action
  implicit none
  type(hmc_para), intent(in) :: para

  c0=ONE;c1=ZERO;c2=ZERO;c3=ZERO
  g2=6.0_8/para%beta
end

!-------------------------------------------------------------------------------
subroutine coefficient_tree(para)

  use typedef_hmc
  use module_action
  implicit none
  type(hmc_para), intent(in) :: para

!!  c0=5.0_8/3.0_8;c1=-ONE/12.0_8;c2=ZERO;c3=ZERO
!!  g2=6.0_8/para%beta
  c0=ONE;c1=-ONE/20.0_8;c2=ZERO;c3=ZERO
  g2=10.0_8/para%beta
end

!-------------------------------------------------------------------------------
subroutine coefficient_iwasaki(para)

  use typedef_hmc
  use module_action
  implicit none
  type(hmc_para), intent(in) :: para

  c0=3.648_8;c1=-0.331_8;c2=ZERO;c3=ZERO
  g2=6.0_8/para%beta
end

!-------------------------------------------------------------------------------
subroutine coefficient_treetad(para)
  use module_input
  use typedef_hmc
  use module_action
  use module_switches
  implicit none
  type(hmc_para), intent(in) :: para
  REAL :: u2

  if (switches%fixtad .and. (.not. switches%fixtad_u04 )) then
     read(input%c0(1)             ,*) c0
     read(input%c1(1)             ,*) c1
     read(input%c2(1)             ,*) c2
     read(input%c3(1)             ,*) c3
     if (c0 /= ONE ) call die("tree tadpole but input%c0 /= 0")
     if (c1 == ZERO) call die("tree tadpole but input%c1 == 0")
     if (c2 /= ZERO) call die("tree tadpole but input%c2 /= 0")
     if (c3 /= ZERO) call die("tree tadpole but input%c3 /= 0")
  else
     u2 = sqrt(plaq_ave)
     g2 = (6.0_8/para%beta) * ( u2 / ( u2 - 0.4_8) )
     c0=ONE;c1=-ONE/20.0_8/u2;c2=ZERO;c3=ZERO
  endif
end

!-------------------------------------------------------------------------------
REAL function csw1_alpha(alpha)

  implicit none
  REAL, intent(in) :: alpha

!  csw1 = 0.00706281_8 + 1.142004_8 * para%alpha &
!                      - 4.194470_8 * para%alpha**2
!  csw1_alpha = 0.00823_8 + 1.13745_8 * alpha &
!                         - 4.18029_8 * alpha**2
   csw1_alpha = 0.19624370857181914_8 &
              + 1.1374515616446979_8 * alpha &
              - 4.1802905917395945_8 * alpha**2



! 0.268185958818892985, if alpha = 0.1.
!
end

!-------------------------------------------------------------------------------
subroutine plaquette_hist(plaq)    ! this is called in hmc
  use module_action
  use module_counter
  implicit none

  REAL, intent(in) :: plaq
  REAL :: diff1, diff2

  DEBUG2S("Start: plaquette_hist")

  if (.not. associated(plaq_hist))call die("ERROR: plaq_hist not associated")

  DEBUG2I("plaquette_hist: counter%traj",counter%traj)
  DEBUG2I("plaquette_hist: ubound plaq_hist",ubound(plaq_hist,1))
  DEBUG2I("plaquette_hist: lbound plaq_hist",lbound(plaq_hist,1))

  if ( counter%traj > ubound(plaq_hist,1) .or. &
       counter%traj < lbound(plaq_hist,1) ) then
     call die("ERROR: out of bound, plaq_hist")
  endif

  plaq_hist(counter%traj) = plaq

  if (counter%traj - plaq_start > 20 )then
     diff1 = plaq_hist(counter%traj  ) - plaq_hist(counter%traj-1)
     diff2 = plaq_hist(counter%traj-1) - plaq_hist(counter%traj-2)
     if ( diff1 * diff2  < 0 ) then
         last_n_ave_plaq = last_n_ave_plaq + 1
     endif
  endif

  n_ave_plaq(counter%traj) = last_n_ave_plaq

!  if (counter%traj-1 == plaq_start) then
!     n_ave_plaq(counter%traj-1) = 1
!  else
!     n_ave_plaq(counter%traj-1) = n_ave_plaq(counter%traj-2)
!  endif

!  if (counter%traj -1 - plaq_start > 20 )then
!     diff1 = plaq_hist(counter%traj-1) - plaq_hist(counter%traj-2)
!     diff2 = plaq_hist(counter%traj-2) - plaq_hist(counter%traj-3)
!     if ( diff1 * diff2  < 0 ) then
!         n_ave_plaq(counter%traj-1) = n_ave_plaq(counter%traj-1) + 1
!     endif
!  endif

  DEBUG2S("End: plaquette_hist")

end

!-------------------------------------------------------------------------------
subroutine read_plaq(para)
  use module_input
  use module_field
  use typedef_para
  use module_counter
  use module_action
  use module_conf_info
  use module_function_decl
  use module_switches
  implicit none

  character(len = 2) :: fmt_name     = "(a"
  character(len = 2) :: fmt_ext      = "a)"
  character(len = 5)  :: fmt_sep      = ",'.',"
  character(len = 4)  :: fmt_run      = "i3.3"

  type(type_para), intent(inout) :: para
  FILENAME_FORMAT :: fmt
  integer :: i, j, iostat
  REAL :: plaq
  REAL, external :: sg

  fmt = fmt_name // fmt_sep // fmt_run // fmt_sep // fmt_ext
  write(plaq_file, fmt)"bqcd" ,para%run, "pinfo"

  if (para%start == START_HOT .or. para%start == START_COLD) then
     plaq_start = - para%nforce
     allocate( plaq_hist(plaq_start : - para%nforce + para%ntraj ))
     allocate(n_ave_plaq(plaq_start : - para%nforce + para%ntraj ))
     n_ave_plaq(plaq_start) = 1
     last_n_ave_plaq = 1
     if (para%start == START_COLD )plaq_hist(plaq_start) = ONE
     if (para%start == START_HOT  )plaq_hist(plaq_start) = ONE !ZERO
  endif

  if (para%start == START_CONT) then
     open(40, file = plaq_file, action = "read", status = "old")
     read(40, *, iostat = iostat) plaq_start ; rewind(40)
     allocate( plaq_hist(plaq_start : counter%traj + para%ntraj ))
     allocate(n_ave_plaq(plaq_start : counter%traj + para%ntraj ))
     iostat = 0 ; j = plaq_start
     do while (iostat == 0)
        read(40, *, iostat = iostat) i, n_ave_plaq(j), plaq_hist(j)
        j = j + 1
     enddo
     close(40)
     last_n_ave_plaq = n_ave_plaq(counter%traj)
  endif

  if (para%start == START_FILE) then
     if (input%io_restart_format == "bqcd") then
        open(40, file=para%info_file(1), action = "read", status = "old")
        call read_keyword_REAL(40, k_plaq, plaq, 1)
        close(40)
     else
        plaq = sg(gauge(1)%u)
     endif

     plaq_start = - para%nforce
     allocate( plaq_hist(plaq_start : - para%nforce + para%ntraj ))
     allocate(n_ave_plaq(plaq_start : - para%nforce + para%ntraj ))
     plaq_hist(plaq_start) = ONE - plaq
     n_ave_plaq(plaq_start) = 1
     last_n_ave_plaq = 1
  endif

  if (gauge_action == "WILSON") then
     switches%improved_gauge = .false.
  else
     switches%improved_gauge = .true.
  endif


  do i = 1, para%n_temper
     call coefficient_sig(para%hmc(i))
!     write(para%c_hmc(i)%c0, "(f20.15)") c0
!     write(para%c_hmc(i)%c1, "(f20.15)") c1
!     write(para%c_hmc(i)%c2, "(f20.15)") c2
!     write(para%c_hmc(i)%c3, "(f20.15)") c3
     write(para%c_hmc(i)%csw, "(f20.15)") para%hmc(i)%csw
     write(para%c_hmc(i)%csw_kappa, "(f20.15)") para%hmc(i)%csw_kappa
     write(para%c_hmc(i)%csw_kappa_strange, "(f20.15)") &
             para%hmc(i)%csw_kappa_strange
  enddo

end

!-------------------------------------------------------------------------------
subroutine write_plaq()

  use module_action
  use module_counter
  use module_function_decl
  implicit none

  integer :: i

  open(40, file = plaq_file, action = "write", status = "unknown")
  do i = plaq_start, counter%traj
     if (my_pe() == 0) write(40, 100)i, n_ave_plaq(i), plaq_hist(i)
  enddo
  close(40)

100 format (1x, 2i10, g30.15)

end

!===============================================================================
