!===============================================================================
!
! dsg.F90  -  p(j,x,mu) := p(j,x,mu) - step * D_{x,mu,j} S_g
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 1998-2006 Hinnerk Stueben
!               2007 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine dsg(p, plaq, step)
  use module_field
  use module_action_type
  use module_vol
  use module_staple
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  type(type_plaq), intent(in)    :: plaq
  REAL,            intent(in)    :: step
  REAL                           :: s
  SU3                            :: uuu, w
  integer                        :: mu, eo, i

  if (plaq%beta == ZERO) return
  if (step == ZERO) return

  TIMING_START(timing_bin_dsg)
  DEBUG2S("Start: dsg")

  call hmc_forces_old(p)

  s = -step * plaq%beta / THREE

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do private(uuu, w)
        do i = 1, volh
           call staple_p(uuu, gauge(1)%u, i, eo, mu)
           call uu(w, gauge(1)%u(1, 1, i, eo, mu), uuu)
           call im_tr_j(p(1, i, eo, mu), w, s)
        enddo
     enddo
  enddo

  call nonactive(p) !! for ddhmc

  call hmc_forces_new(p, step, plaq%fid)

  DEBUG2S("End: dsg")
  TIMING_STOP(timing_bin_dsg)
end

!===============================================================================
