!===============================================================================
!
! hmc.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 1998-2007 Hinnerk Stueben
!               2007-2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine gauge_copy(out, in)
  use module_vol
  implicit none
  GAUGE_FIELD :: out, in
  integer :: mu, eo, i

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do
        do i = 1, volh_tot
           out(1, 1, i, eo, mu) = in(1, 1, i, eo, mu)
           out(2, 1, i, eo, mu) = in(2, 1, i, eo, mu)
           out(3, 1, i, eo, mu) = in(3, 1, i, eo, mu)
           out(1, 2, i, eo, mu) = in(1, 2, i, eo, mu)
           out(2, 2, i, eo, mu) = in(2, 2, i, eo, mu)
           out(3, 2, i, eo, mu) = in(3, 2, i, eo, mu)
           out(1, 3, i, eo, mu) = in(1, 3, i, eo, mu)
           out(2, 3, i, eo, mu) = in(2, 3, i, eo, mu)
           out(3, 3, i, eo, mu) = in(3, 3, i, eo, mu)
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine hmc(para, conf, out, force_accept, test)
  use module_stout_field
  use module_field
  use module_cg
  use typedef_hmc
  use module_sf
  use module_counter
  use module_function_decl
  use module_p_interface
  use module_switches
  use module_vol
  use module_input
  implicit none

  type(hmc_para), intent(in)    :: para
  type(hmc_conf), intent(inout) :: conf
  type(hmc_out),  intent(out)   :: out
  integer,        intent(in)    :: force_accept
  integer,        intent(in)    :: test

  P_GAUGE_FIELD, save         :: u_bck
!!  P_SPINCOL_FIELD,  save      :: phi_bck
  P_GENERATOR_FIELD, save     :: p
  integer :: i

  REAL                        :: sd_old, sd_new, sd_dif
  REAL                        :: sf1_old, sf1_new, sf1_dif
  REAL                        :: sf2_old, sf2_new, sf2_dif
  REAL                        :: sf3_old, sf3_new, sf3_dif
  REAL                        :: sg_old, sg_new, sg_dif
  REAL                        :: sp_old, sp_new, sp_dif
  REAL                        :: hg_old, hg_new, hg_dif
  REAL                        :: h_old, h_new, h_dif
  REAL :: plaq, plaq_old(3), plaq_new(3), rect_old(3), rect_new(3)
  REAL, external              :: sp, sig, sg

  DEBUG2S("Start HMC")
  TIMING_START(timing_bin_hmc)
  TIMING_START(timing_bin_hmc_init)

  ! initialize clover terms and stout links
  call coefficient_sig(para)
  call init_imp_all(.false.)
  if (switches%measurement_only) then
     out%sg = sg(gauge(1)%u) !!! ; call check_stout();stop
     return
  endif

  hg_old = ZERO;  hg_new = ZERO
  sg_old = ZERO;  sg_new = ZERO
  sp_old = ZERO;  sp_new = ZERO
  sd_old = ZERO;  sd_new = ZERO
  sf1_old = ZERO;  sf1_new = ZERO
  sf2_old = ZERO;  sf2_new = ZERO
  sf3_old = ZERO;  sf3_new = ZERO
 
  call init_cg_stat(cg_stat)
  call init_cg_stat(cgm_stat)
  call init_cg_stat(cg_ritz_stat)
  call init_cg_stat2(dd_stat)
  call init_cg_stat2(hh_stat)
  call iteration_count_init()

  call init_hmc(para)
  call init_rhmc(para)

  if (.not. associated(p)) then
  nullify(p)
  nullify(u_bck)
!!  nullify(phi_bck)
                          ALLOCATE_GEN_FIELD(p)
                          ALLOCATE_G_FIELD(u_bck)
!!                          ALLOCATE_SC_FIELD(phi_bck)
  endif


  TIMING_STOP(timing_bin_hmc_init)

  if (test /= HMC_TEST_BACKWARDS) then  ! ie normally do:

     ! backups:
     TIMING_START(timing_bin_hmc_backup)
!!     if (input%hmc_mpf_mass /= 0 )call swap_p_sc_field(phi_bck, mpf(1)%sc)
  DEBUG2S("HMC: u_bck")
     call gauge_copy(u_bck, gauge(1)%u)
  DEBUG2S("HMC: u_bck done")
     TIMING_STOP(timing_bin_hmc_backup)

     TIMING_START(timing_bin_hmc_h_old)

     ! initialize momenta p, phi, phi2 and old action:
#ifdef ZERO_MOMENTUM
     p = 0
#else
     call hmc_init_p(p)
#endif

     call hmc_init_phi(sf1_old, sf2_old, sf3_old)

     sp_old = sp(p)
     hg_old = sig(sg_old, plaq_old, rect_old)
     call clover_action_sum(sd_old)
     h_old = sd_old + sp_old + hg_old + sf1_old + sf2_old + sf3_old
     TIMING_STOP(timing_bin_hmc_h_old)
  endif
  if (test == HMC_TEST_FORWARDS) then
     call hmc_test_report(test, p, gauge(1)%u, &
                          sp_old, hg_old, sf1_old, sf2_old+sf3_old, sd_old)
  endif


!--------------------------------------------------------------------- MD step:
  call integrate(p, para, conf, sf1_new, sf2_new, sf3_new)
!------------------------------------------------------- calculate Hamiltonian:
  DEBUG2S("Start calculate Hamiltonian")
  TIMING_START(timing_bin_hmc_h_new)
  sp_new = sp(p)
  hg_new = sig(sg_new, plaq_new, rect_new)
  call clover_action_sum(sd_new)
  h_new = sd_new + sp_new + hg_new + sf1_new + sf2_new + sf3_new
  TIMING_STOP(timing_bin_hmc_h_new)
  DEBUG2S("End calculate Hamiltonian")
!------------------------------------------------------------- accept new U ? :


  TIMING_START(timing_bin_hmc_rest)
  DEBUG2S("Start MC check")
  hg_dif =   hg_new  -  hg_old
  sp_dif =   sp_new  -  sp_old
  sd_dif =   sd_new  -  sd_old
  sf1_dif =  sf1_new -  sf1_old
  sf2_dif =  sf2_new -  sf2_old
  sf3_dif =  sf3_new -  sf3_old
  h_dif = sd_dif + sp_dif + hg_dif + sf1_dif + sf2_dif + sf3_dif
!!  out%exp_dh = exp(h_old - h_new)
  out%exp_dh = exp( - h_dif)

  if (force_accept /= 0) then
     out%accepted = 1
  else
     if (ran_number() < out%exp_dh) then
        out%accepted = 1
     else
        out%accepted = 0
     endif
  endif
  DEBUG2S("End MC check")

  if (out%accepted == 1) then
     out%sg = sg_new
     out%sf = sf1_new
     out%plaq = plaq_new
     out%rect = rect_new
  else
     out%plaq = plaq_old
     out%rect = rect_old
     DEBUG2S("Start replace new by old fields")
!!     if (input%hmc_mpf_mass /= 0 ) call swap_p_sc_field(mpf(1)%sc, phi_bck)
  DEBUG2S("HMC: u_bck")
     call gauge_copy(gauge(1)%u, u_bck)
  DEBUG2S("HMC: u_bck done")
!!     gauge(1)%u(1:3,1:3,1:volh_tot,0:1,1:4) = u_bck(1:3,1:3,1:volh_tot,0:1,1:4)
     gauge(1:size(gauge))%init = .true.
     if(associated(stout)) stout(1:size(stout))%init = .true.
     if(associated(clover))clover(1:size(clover))%init = .true.
     DEBUG2S("End replace new by old fields")
  endif

  call paratransu(gauge(1)%u) !! need for nonactive link method for DD


  DEBUG2S("Record CG stat")
  call get_cg_stat(cg_stat,  out%cg_ncall, out%cg_niter_max, out%cg_niter_tot)
  call get_cg_stat(cgm_stat, out%cgm_ncall, out%cgm_niter_max, out%cgm_niter_tot)

  DEBUG2S("Write iterCount and Force")
  call iteration_count_is_write(UREC)
  call hmc_forces_write(UREC)
  call rhmc_forces_write(UREC)

  if (test == HMC_TEST_BACKWARDS) then
     call hmc_test_report(test, p, gauge(1)%u, &
                          sp_new, hg_new,sf1_new, sf2_new+sf3_new, sd_new)
  endif
  TIMING_STOP(timing_bin_hmc_rest)

  DEBUG2S("Write Action old new diff")
  call hmc_write_h(UREC, "old",  h_old, sp_old, hg_old, sd_old, &
  sf1_old, sf2_old, sf3_old)
  call hmc_write_h(UREC, "new",  h_new, sp_new, hg_new, sd_new, &
  sf1_new, sf2_new, sf3_new)
  call hmc_write_h(UREC, "diff", h_dif, sp_dif, hg_dif, sd_dif, &
  sf1_dif, sf2_dif, sf3_dif)

  TIMING_STOP(timing_bin_hmc)

  if (switches%stout .and. para%n_stout == 5) then
     DEBUG2S("Check stout")
     call check_stout(conf)
  endif

  plaq = ONE - out%sg / (SIX * volume)
  call plaquette_hist(plaq)

  DEBUG2S("End HMC")

end

!===============================================================================
