!===============================================================================
!
! hmc_integrator_p.F90 - integrators for the different models
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine hmc_integrator_p(p, para, conf, i, calc_sf)
  use module_integrator
  use typedef_hmc
  use module_vol
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  type(hmc_para),  intent(in)    :: para
  type(hmc_conf),  intent(in)    :: conf
  integer,         intent(in)    :: calc_sf, i
  integer :: j

  do j = 1, n_time_scale
     call hmc_inte_p_control(p, para, conf, step_q(i,j), calc_sf, j)
  enddo

end

!-------------------------------------------------------------------------------
subroutine hmc_inte_p_control(p, para, conf, step, calc_sf, i)
  use module_field
  use module_action

  use module_input
  use typedef_hmc
  use module_sf
  use module_vol
  use module_rhmc
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  type(hmc_para),  intent(in)    :: para
  type(hmc_conf),  intent(in)    :: conf
  REAL,            intent(in)    :: step
  integer,         intent(in)    :: calc_sf, i
  integer :: j, p_done, acmpf, top_rhmc
  integer, allocatable :: shift_switch(:)
  REAL,    allocatable :: shift_step(:)
  character(len = 80) :: msg

  if (step == ZERO .and. calc_sf == 0) return

  DEBUG2S("Start: hmc_inte_p_control")
  TIMING_START(timing_bin_hmc_steps)

  p_done = 0

!!  if (input%hmc_dsf_hhdd_test == i) then
!!     acmpf = accumulated_num_mpf(HH_NORMAL_DD)+1
!!     call dsf_hhdd(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
!!  endif
!!  if (input%hmc_dsf_hhbb_test == i) then
!!     acmpf = accumulated_num_mpf(HH_NORMAL_BB)+1
!!     call dsf_hhbb(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
!!  endif

  if (input%hmc_dsf1_un == i) then
     acmpf = accumulated_num_mpf(UN_NORMAL)+1
     call dsf_unprec_wilson(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif
  if (input%hmc_dsf2_un == i) then
     acmpf = accumulated_num_mpf(UN_NORMAL)+2
     call dsf_unprec_wilson(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif

  if (input%hmc_dsf1_dd == i) then
     acmpf = accumulated_num_mpf(DD_NORMAL)+1
     call dsf_dd(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif
  if (input%hmc_dsf2_dd == i) then
     acmpf = accumulated_num_mpf(DD_NORMAL)+2
     call dsf_dd(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif

  if (input%hmc_dsf1_hh == i) then
     acmpf = accumulated_num_mpf(HH_NORMAL)+1
     call dsf_hh(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif
  if (input%hmc_dsf2_hh == i) then
     acmpf = accumulated_num_mpf(HH_NORMAL)+2
     call dsf_hh(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif

  if (input%hmc_dsf1_mtmp == i) then
     acmpf = accumulated_num_mpf(EO_NORMAL)+1
     call dsf_mtmp(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif
  if (input%hmc_dsf2_mtmp == i) then
     acmpf = accumulated_num_mpf(EO_NORMAL)+2
     call dsf_mtmp(p, step, calc_sf, sf(acmpf), acmpf);p_done=p_done+1
  endif



  if (input%hmc_dsf_erl == i) then
     acmpf = accumulated_num_mpf(EO_RATIONAL)+1
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif
  if (input%hmc_dsf_ers == i) then
     acmpf = accumulated_num_mpf(EO_RATIONAL)+1 + input%hmc_mpf_rhmc_l
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif

  if (input%hmc_dsf_drl == i) then
     acmpf = accumulated_num_mpf(DD_RATIONAL)+1
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif

  if (input%hmc_dsf_drs == i) then
     acmpf = accumulated_num_mpf(DD_RATIONAL)+1 + input%hmc_mpf_dd_rhmc_l
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif

  if (input%hmc_dsf_hrl == i) then
     acmpf = accumulated_num_mpf(HH_RATIONAL)+1
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif
  if (input%hmc_dsf_hrs == i) then
     acmpf = accumulated_num_mpf(HH_RATIONAL)+1 + input%hmc_mpf_hh_rhmc_l
     allocate(shift_switch(ratapp(0)%shift));shift_switch=0
     allocate(shift_step(ratapp(0)%shift))  ;shift_step=step
     call dsfr(shift_switch, p, shift_step, calc_sf, acmpf);p_done = p_done + 1
     deallocate(shift_switch)
     deallocate(shift_step)
  endif


!!  if (input%hmc_dsfr == i) then
!!     allocate(shift_switch(shift))
!!     allocate(shift_step(shift))
!!     shift_switch=0
!!     shift_step=step
!!!     do j = 1 , shift
!!!        select case (time_switch(j))
!!!           case(1); shift_step(j)=step
!!!           case(2); shift_step(j)=step / para%m_scale
!!!        end select
!!!     enddo
!!     call dsfr(shift_switch, p, conf, shift_step, calc_sf, 1, para)
!!     p_done = p_done + 1
!!     deallocate(shift_switch)
!!     deallocate(shift_step)
!!  endif
!!!  if (input%hmc_dsfr + 1 == i .and. 0 < im  .and. im < para%m_scale) then
!!!     do i=1,shift
!!!        select case (time_switch(i))
!!!           case(1); shift_switch(i)=1; shift_step(i)=ZERO
!!!           case(2); shift_switch(i)=0; shift_step(i)=step
!!!        end select
!!!     enddo
!!!     call dsfr(shift_switch, p, conf, shift_step, calc_sf, sfr, 2, para)
!!!     p_done = p_done + 1
!!!  endif

  if (input%hmc_dsd == i) then
     call dsd(p, step)       ; p_done = p_done + 1
  endif
  if (input%hmc_dsg == i) then
     call dsg(p, action%plaq, step) ; p_done = p_done + 1
  endif
  if (input%hmc_dsig == i) then
     call dsig(p, action%impg, step); p_done = p_done + 1
  endif

  if (p_done == 0) then
     msg = "hmc_inte_p_control : didn't anything at scale " // char(48+i)
     call warn(msg)
     call die(msg)
  endif

  if (calc_sf == 1) then
     top_rhmc=999
     if (input%hmc_dsf_erl/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_erl)
     if (input%hmc_dsf_drl/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_drl)
     if (input%hmc_dsf_hrl/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_hrl)
     if (input%hmc_dsf_ers/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_ers)
     if (input%hmc_dsf_drs/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_drs)
     if (input%hmc_dsf_hrs/=0)top_rhmc=min(top_rhmc,input%hmc_dsf_hrs)
     if (i==top_rhmc)call calc_fermi_action_rhmc()
  endif

  TIMING_STOP(timing_bin_hmc_steps)
  DEBUG2S("End: hmc_inte_p_control")
end

!===============================================================================
