!===============================================================================
!
! hmc_u.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 1998-2007 Hinnerk Stueben
!               2007-2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine hmc_integrator_q(p, para, conf, step)
  use module_field
  use module_stout_field

  use typedef_hmc
  use module_switches
  use module_vol

  use module_dd_evec
  implicit none

  GENERATOR_FIELD, intent(in)    :: p
  type(hmc_para),  intent(in)    :: para
  type(hmc_conf),  intent(inout) :: conf
  REAL,            intent(in)    :: step

  if (step == ZERO) return

  call hmc_u(p, conf, step, para)

!!  if (switches%boundary_sf) call schr_boundary_gauge(gauged(1)%u)

  gauge(1:size(gauge))%init = .true.
  if (associated(stout)) stout(1:size(stout))%init = .true.
  if (associated(clover))clover(1:size(clover))%init = .true.
  if (dd_evec   ) evecdd(   :,:)%fresh=.false.
  if (dd_evec_r4) evecdd_r4(:,:)%fresh=.false.
end
!-------------------------------------------------------------------------------
subroutine hmc_u(p, conf, step, para)

  use module_field
  use typedef_hmc
  use module_vol
  implicit none

  GENERATOR_FIELD, intent(in)    :: p
  type(hmc_conf),  intent(inout) :: conf
  REAL,            intent(in)    :: step
  type(hmc_para),  intent(in)    :: para

  GENERATOR                      :: q
  SU3                            :: v,q1,q2,f,b1,b2
  integer                        :: i, mu, eo, j

  DEBUG2S("Start: hmc_u")
  TIMING_START(timing_bin_hmc_u)

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do private(j, q, v, q1,q2,f,b1,b2)
        do i = 1, VOLH
           do j = 1, NGEN
              q(j) = p(j, i, eo, mu) * step
           enddo
!           call gen2u(v, q)

           call gen2tlhermit(q1,q)
           call projection_su3(q1, v, q2, b1, b2, f, .false.)
           call u_update(gauge(1)%u(:, :, i, eo, mu), v)  ! u = v * u
           call u_normalize(gauge(1)%u(:, :, i, eo, mu))
        enddo
     enddo
  enddo
  
  TIMING_STOP(timing_bin_hmc_u)


  TIMING_START(timing_bin_hmc_xbound_g)
!!  call xbound_g_field(gauge(1)%u(1,1,1,0,1))
  call xbound_g_field(gauge(1)%u)
  TIMING_STOP(timing_bin_hmc_xbound_g)

  DEBUG2S("End: hmc_u")
end

!===============================================================================
