!===============================================================================
!
! init_hmc.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_hmc(para)
  use typedef_hmc
  use module_field
  use module_input
  use module_sf
  use module_integrator
  implicit none
  type(hmc_para), intent(in)    :: para


  if (.not. associated(sf  ) .and. associated(mpf)) allocate(sf( size(mpf) ))

  n_time_scale = 6
!  if(para%m_scale5 == 1) n_time_scale = 5
!  if(para%m_scale4 == 1) n_time_scale = 4
!  if(para%m_scale3 == 1) n_time_scale = 3
!  if(para%m_scale2 == 1) n_time_scale = 2
!  if(para%m_scale  == 1) n_time_scale = 1

  if(para%m_scale5 == 0 .or. (para%m_scale5==1 .and. para%integrator(6)=="LPFSTS") .or. para%integrator(6)=="NON") n_time_scale = 5
  if(para%m_scale4 == 0 .or. (para%m_scale4==1 .and. para%integrator(5)=="LPFSTS") .or. para%integrator(5)=="NON") n_time_scale = 4
  if(para%m_scale3 == 0 .or. (para%m_scale3==1 .and. para%integrator(4)=="LPFSTS") .or. para%integrator(4)=="NON") n_time_scale = 3
  if(para%m_scale2 == 0 .or. (para%m_scale2==1 .and. para%integrator(3)=="LPFSTS") .or. para%integrator(3)=="NON") n_time_scale = 2
  if(para%m_scale  == 0 .or. (para%m_scale ==1 .and. para%integrator(2)=="LPFSTS") .or. para%integrator(2)=="NON") n_time_scale = 1

  if(input%hmc_dsf1_un           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf2_un           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf1_dd           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf2_dd           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf1_hh           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf2_hh           >n_time_scale)call die("ds: out of integrator")

  if(input%hmc_dsf1_mtmp         >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf2_mtmp         >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf3_mtmp         >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf4_mtmp         >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf5_mtmp         >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_erl           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_drl           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_hrl           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_ers           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_drs           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsf_hrs           >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsd               >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsg               >n_time_scale)call die("ds: out of integrator")
  if(input%hmc_dsig              >n_time_scale)call die("ds: out of integrator")




  if (.not. associated(m_scale))allocate(m_scale(n_time_scale ))
  if (.not. associated(scale  ))allocate(scale(  n_time_scale ))
  if (.not. associated(jscale ))allocate(jscale( n_time_scale ))
  if (.not. associated(step   ))allocate(step(   n_time_scale ))

                        m_scale(1) = para%ntau
  if (n_time_scale > 1) m_scale(2) = para%m_scale
  if (n_time_scale > 2) m_scale(3) = para%m_scale2
  if (n_time_scale > 3) m_scale(4) = para%m_scale3
  if (n_time_scale > 4) m_scale(5) = para%m_scale4
  if (n_time_scale > 5) m_scale(6) = para%m_scale5

                        scale(1) = para%ntau
  if (n_time_scale > 1) scale(2) = scale(1) * para%m_scale
  if (n_time_scale > 2) scale(3) = scale(2) * para%m_scale2
  if (n_time_scale > 3) scale(4) = scale(3) * para%m_scale3
  if (n_time_scale > 4) scale(5) = scale(4) * para%m_scale4
  if (n_time_scale > 5) scale(6) = scale(5) * para%m_scale5

  if (.not. associated(integrator))allocate(integrator(n_time_scale))

  integrator(1:n_time_scale) = para%integrator(1:n_time_scale)
end

!-------------------------------------------------------------------------------
integer function num_time_scale(para)
  use typedef_hmc
  use module_input
  implicit none
  type(hmc_para), intent(in) :: para

  num_time_scale = 6
  if(para%m_scale5 == 0 .or. (para%m_scale5==1 .and. para%integrator(6)=="LPFSTS") .or. para%integrator(6)=="NON") num_time_scale = 5
  if(para%m_scale4 == 0 .or. (para%m_scale4==1 .and. para%integrator(5)=="LPFSTS") .or. para%integrator(5)=="NON") num_time_scale = 4
  if(para%m_scale3 == 0 .or. (para%m_scale3==1 .and. para%integrator(4)=="LPFSTS") .or. para%integrator(4)=="NON") num_time_scale = 3
  if(para%m_scale2 == 0 .or. (para%m_scale2==1 .and. para%integrator(3)=="LPFSTS") .or. para%integrator(3)=="NON") num_time_scale = 2
  if(para%m_scale  == 0 .or. (para%m_scale ==1 .and. para%integrator(2)=="LPFSTS") .or. para%integrator(2)=="NON") num_time_scale = 1
end
!===============================================================================
