!===============================================================================
!
! chemicalp.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine apply_mp_chemu(oe, oo, ie, io, mid, chemu, i)! d^i M / (d mu)^i
  use module_field
  use module_vol
  use module_action
  implicit none
  REAL, intent(in) :: chemu
  integer, intent(in) :: mid, i
  SPINCOL_FIELD, intent(out) :: oe, oo
  SPINCOL_FIELD, intent(in)  :: ie, io
  REAL :: fac1, fac2

  if (i <= 0) call die("apply_mp_chemu: i<=0")

  fac1 = exp( chemu)
  if (mod(i,2)==0) then
     fac2 = exp(-chemu)
  else
     fac2 =-exp(-chemu)
  endif

  call emuu4(gauge(2)%u, fac1, fac2)
  call d(ODD, EVEN, oo, ie, gauge(2)%u)
  call d(EVEN, ODD, oe, io, gauge(2)%u)
  oo = - action%mtilde(mid)%kappa * oo
  oe = - action%mtilde(mid)%kappa * oe
  call d_g_u_reorder(gauge(2)%u)

end

!-------------------------------------------------------------------------------
subroutine emuu4(u, fac1, fac2)
  use module_d_g_reorder
  use module_nn
  use module_vol
  use module_lattice
  implicit none
  GAUGE_FIELD, intent(in) :: u
  REAL, intent(in) :: fac1, fac2
  integer :: eo, oe, i, j, mu, col, c1, c2

  u_reorder = 0
  mu = gamma_index(4)

  do eo = EVEN, ODD
     oe = 1 - eo
     !$omp parallel do private(j, col)
     do i = 1, volh
        j=nn(i,eo,mu,BWD)
        do col = 1, NCOL
           u_reorder(1, col, i, mu, eo) =       u(1, col, i, eo, mu)  *fac1
           u_reorder(2, col, i, mu, eo) =       u(2, col, i, eo, mu)  *fac1
           u_reorder(3, col, i, mu, eo) =       u(3, col, i, eo, mu)  *fac1
           u_reorder(4, col, i, mu, eo) = conjg(u(col, 1, j, oe, mu)) *fac2
           u_reorder(5, col, i, mu, eo) = conjg(u(col, 2, j, oe, mu)) *fac2
           u_reorder(6, col, i, mu, eo) = conjg(u(col, 3, j, oe, mu)) *fac2
        enddo
     enddo
  enddo

end

!===============================================================================
