!===============================================================================
!
! correlation_p5p.F90 - correlation psi_gamma5_psi and psi_gamma5_psi
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"
# include "defs_imp_2p1.h"

!-------------------------------------------------------------------------------
subroutine correlation_p5p(c_t0t1, &
                           psi1_e,psi2_e,eta1_e,eta2_e,&
                           psi1_o,psi2_o,eta1_o,eta2_o)

  use module_vol
  implicit none

  COMPLEX, dimension(0:LT-2,0:1),intent(out):: c_t0t1
  SPINCOL_FIELD, intent(in)             :: psi1_e, psi2_e, psi1_o, psi2_o
  SPINCOL_FIELD, intent(in)             :: eta1_e, eta2_e, eta1_o, eta2_o
  COMPLEX,dimension(volh) :: psi_e, psi_o, eta_e, eta_o
  COMPLEX,dimension(vol ) :: psi, eta    ! <--------std
  COMPLEX, external :: correlation_tt
  COMPLEX :: c
  integer :: lt, t0, dis_t0_t1, std_t1, coord(DIM)

  lt = decomp%std%L(4)
  nt = decomp%std%N(4)

  call real_coord(coord)
  c_t0t1 = ZERO

  call psi_g5_psi(psi_e, psi1_e, psi2_e)
  call psi_g5_psi(psi_o, psi1_o, psi2_o)
  call psi_g5_psi(eta_e, eta1_e, eta2_e)
  call psi_g5_psi(eta_o, eta1_o, eta2_o)
  call reduction_eo_std(psi, psi_e, psi_o)
  call reduction_eo_std(eta, eta_e, eta_o)


  do dd = 0, lt - 2
  do it = 0, lt - 1
     jt = it + dd
     do ix = 0, lx - 1
     do iy = 0, ly - 1
     do iz = 0, lz - 1
        do jx = 0, lx - 1
        do jy = 0, ly - 1
        do jz = 0, lz - 1

        in1(ix,iy,iz,it)*in1(jx,jy,jz,jt)






  do dis_t0_t1 = 0, lt - 2
     do t0 = 0, nt - 1


        std_t1 = coord(4) + t0 + dis_t0_t1
        c = correlation_tt(t0, dis_t0_t1, psi, eta)
        if (std_t1 < L(4)) then
           c_t0t1(dis_t0_t1,0) = c_t0t1(dis_t0_t1,0) + c
        else
           c_t0t1(dis_t0_t1,1) = c_t0t1(dis_t0_t1,1) + c
        endif
     enddo
     c_t0t1(dis_t0_t1,0) = c_t0t1(dis_t0_t1,0) / dble( LT - dis_t0_t1 )
     if (dis_t0_t1 /= 0) then
        c_t0t1(dis_t0_t1,1) = c_t0t1(dis_t0_t1,1) / dble( dis_t0_t1 )
     endif
  enddo

end

!-------------------------------------------------------------------------------
subroutine psi_g5_psi(p5p,psi1,psi2)

  use module_vol
  use module_p_interface
  implicit none
  SPINCOL_FIELD, intent(in)             :: psi1, psi2
  COMPLEX,dimension(volh),intent(out)   :: p5p
  P_SPINCOL_FIELD, save                 :: tmp
  COMPLEX, external :: cdotc_12
  integer :: i

  ALLOCATE_SC_FIELD(tmp)

  tmp=psi2
  call gamma5(tmp,volh)

  !$omp parallel do
  do i = 1, volh
     p5p(i) = cdotc_12(psi1(1,1,i), tmp(1,1,i))
  enddo

end

!===============================================================================
