!===============================================================================
!
! correlations.F90 - correlation function
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine correlations(para, conf, traj, i_ensemble1, i_ensemble2)
  use module_field
  use module_action

  use module_input
  use typedef_hmc
  use module_lattice
  use module_function_decl
  use module_p_interface
  use module_vol
  use module_switches
  implicit none

  integer,        intent(in)  :: traj, i_ensemble1, i_ensemble2
  type(hmc_para), intent(in)  :: para
  type(hmc_conf), intent(in)  :: conf

  P_SPINCOL_FIELD, save       :: eta_e, eta_o, zeta_e, zeta_o,tmp1,tmp2

  character(len=6) :: name

  character(len=*), parameter :: key_tr = "%tr"
  integer,        save        :: count = 0
  integer                     :: n_sc_field, size_of_trace
  integer                     :: cg_ncall, cg_niter_max, cg_niter_tot
  REAL                        :: pinorm
  REAL                        :: re_pbp, im_pbp, re_p5p, im_p5p
  COMPLEX                     :: pbp, p5p
  REAL                        :: res(5)
  integer :: i, jc  ,x,y,z,t, iterations, fid, mid, cid

  if (.not. switches%measure_correlations) return

  fid = input%hmc_mpf_mass
  mid = action%fermi(fid)%mid1
  cid = action%mtilde(mid)%cid

  ALLOCATE_SC_FIELD(eta_e)
  ALLOCATE_SC_FIELD(eta_o)
  ALLOCATE_SC_FIELD(zeta_e)
  ALLOCATE_SC_FIELD(zeta_o)
  ALLOCATE_SC_FIELD(tmp1)
  ALLOCATE_SC_FIELD(tmp2)

  count = count + 1
  n_sc_field = NDIRAC * NCOL * volh
  size_of_trace = NDIRAC * NCOL * volume

!!  call ran_gauss_volh(NDIRAC * NCOL, eta_e, HALF, EVEN)
!!  call ran_gauss_volh(NDIRAC * NCOL, eta_o, HALF, ODD)

  call init_quark(mid)
  call make_source_0000(eta_e,eta_o)
  call write_vector12_0000_sum(eta_e, eta_o, "sorce ")

  call d(ODD, EVEN, tmp1, eta_e, gauge(2)%u)
  call write_vector12_0000_sum(tmp1, eta_o, "Doe_se")

  call clover_mult_a(tmp1, clover(cid)%a(1,1,EVEN), eta_e, volh)
  call write_vector12_0000_sum(tmp1, eta_o, "cee_se")

  call make_source_0000(eta_e,eta_o)
  call d(ODD, EVEN, tmp1, eta_e, gauge(2)%u)
  call clover_mult_ao(clover(cid)%i(1,1,ODD), tmp1, volh)
  call d(EVEN, ODD, tmp2, tmp1, gauge(2)%u)
  call write_vector12_0000_sum(tmp2, eta_o, "dcd_se")

  call make_source_0000(eta_e,eta_o)
  call mtil(tmp1, eta_e, mid)
#ifdef OMTDTD
  if (action%mtilde(mid)%cswkappa /= ONE) then
     call clover_mult_a(tmp2, clover(cid)%a(1,1,EVEN), tmp1, volh)
  else
     if (action%mtilde(mid)%h /= ONE) &
     call h_mult_a(tmp2, real(action%mtilde(fid)%h,kind=RKIND), tmp1, volh)
  endif
#else
  tmp2 = tmp1
#endif
  call write_vector12_0000_sum(tmp2, eta_o, "mtl_se")

  call make_source_0000(eta_e,eta_o)
!!  call init_cg_stat()
  call solve(mid, zeta_e, zeta_o, eta_e, eta_o)  ! zeta = inv(M) eta
!!  call get_cg_stat(cg_ncall, cg_niter_max, cg_niter_tot)
  call write_vector12_0000_sum(zeta_e, zeta_o, "solver")


  call make_source_0000(eta_e,eta_o)

  do i = 1, 8
     write(name,50)"DDee_",i
     call d(ODD, EVEN, tmp1, eta_e, gauge(2)%u)
     call d(EVEN, ODD, eta_e, tmp1, gauge(2)%u)
     call write_vector12_0000_sum(eta_e, eta_o, name)
  enddo

50 format( a5,i1 )

end

!===============================================================================
