!===============================================================================
!
! traces2.F90
!
! calculates: Tr(inv(M))          psibar psi (pbp)
!             Tr(gamma5 inv(M))   psibar gamma5 psi (p5p)
!             Tr(inv(M+ M))       pion norm  (pinorm)
!
! traces of a matrix A are calculated with a stochastic estimator:
!
!     Tr(A) = eta+ A eta  (eta: Gaussian noise)
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine traces2(traj)
  use typedef
  use module_p_interface
  use module_vol
  use module_cg
  use module_input
  use module_action
  use module_function_decl
  implicit none

  integer,        intent(in)  :: traj
!!  type(sc_field), dimension(:),pointer,save :: srce, srco, rese, reso
  P_SPINCOL_FIELD, save       :: eta_e, eta_o, zeta_e, zeta_o, tmpe, tmpo, tmp1e,tmp1o
  COMPLEX :: tmp, tmp1, tmp2, imdm,imd2m, imdmimdm
  integer :: i, nsrc, fid, mid
  REAL :: cswkappa, invm_sf

  DEBUG2S("Start: traces2")
  fid = input%hmc_mpf_mass
  mid = action%fermi(fid)%mid1
  cswkappa= action%mtilde(mid)%cswkappa

  nsrc = 50

  ALLOCATE_SC_FIELD(eta_e)
  ALLOCATE_SC_FIELD(eta_o)
  ALLOCATE_SC_FIELD(zeta_e)
  ALLOCATE_SC_FIELD(zeta_o)
  ALLOCATE_SC_FIELD(tmpe)
  ALLOCATE_SC_FIELD(tmpo)
  ALLOCATE_SC_FIELD(tmp1e)
  ALLOCATE_SC_FIELD(tmp1o)

!!  allocate(srce(nsrc))
!!  allocate(srco(nsrc))
!!  allocate(rese(nsrc))
!!  allocate(reso(nsrc))
!!  do i=1,nsrc
!!     nullify(srce(i)%sc)
!!     nullify(srco(i)%sc)
!!     nullify(rese(i)%sc)
!!     nullify(reso(i)%sc)
!!     ALLOCATE_SC_FIELD(srce(i)%sc)
!!     ALLOCATE_SC_FIELD(srco(i)%sc)
!!     ALLOCATE_SC_FIELD(rese(i)%sc)
!!     ALLOCATE_SC_FIELD(reso(i)%sc)
!!  enddo
!!  do i=1,nsrc
!!  call ran_gauss_volh(NDIRAC * NCOL, srce(i)%sc, HALF, EVEN)
!!  call ran_gauss_volh(NDIRAC * NCOL, srco(i)%sc, HALF, ODD )
!!  enddo


  invm_sf=0
  imd2m = 0
  imdm = 0
  imdmimdm =0
  do i=1,nsrc
     call ran_gauss_volh(NDIRAC * NCOL, eta_e, HALF, EVEN)
     call ran_gauss_volh(NDIRAC * NCOL, eta_o, HALF, ODD )
     call solve(mid, zeta_e, zeta_o, eta_e, eta_o)


     call unprec_cmul(tmpe, tmpo, zeta_e, zeta_o, mid)
     tmpe = zeta_e - tmpe
     tmpo = zeta_o - tmpo
     tmp = sc_cdotc(eta_e, tmpe) + sc_cdotc(eta_o, tmpo)
     call global_sum_vec(2, tmp)
     invm_sf = invm_sf + tmp


     call apply_mp_chemu(tmpe, tmpo, zeta_e, zeta_o, mid, 0.0_8, 2)
     tmp = sc_cdotc(eta_e, tmpe) + sc_cdotc(eta_o, tmpo)
     call global_sum_vec(2, tmp)
     imd2m = imd2m + tmp

     call apply_mp_chemu(tmpe, tmpo, zeta_e, zeta_o, mid, 0.0_8, 1)
     tmp = sc_cdotc(eta_e, tmpe) + sc_cdotc(eta_o, tmpo)
     call global_sum_vec(2, tmp)
     imdm = imdm + tmp

     call solve(mid, tmp1e, tmp1o, tmpe, tmpo)
     call apply_mp_chemu(tmpe, tmpo, tmp1e, tmp1o, mid, 0.0_8, 1)
     tmp = sc_cdotc(eta_e, tmpe) + sc_cdotc(eta_o, tmpo)
     call global_sum_vec(2, tmp)
     imdmimdm = imdmimdm + tmp

  enddo

  if (my_pe()==0)write(UREC,100)"%tr_SFinvM", invm_sf/dble(nsrc)/cswkappa
  if (my_pe()==0)write(UREC,101)"%tr_invMdM", imdm/dble(nsrc)
  if (my_pe()==0)write(UREC,101)"%tr_invMd2M", imd2m/dble(nsrc)
  if (my_pe()==0)write(UREC,101)"%tr_invMdMinvMdM", imdmimdm/dble(nsrc)

100 format (2x, a, f20.5)
101 format (2x, a, 2f20.5)
  DEBUG2S("End: traces2")

400 format (1x, 2a, a6, 2a3, 5a20,    a10)
410 format (1x, a4, i6, 2i3, 5g20.10, i10)

end

!===============================================================================
