!===============================================================================
!
! vector12.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine global_sum_vector12(sc,in_e,in_o)

  use module_function_decl
  use module_vol
  implicit none
  COMPLEX, dimension(DIM,NCOL), intent(out) :: sc
  SPINCOL_FIELD, intent(in) :: in_e, in_o
  COMPLEX :: tmp
  REAL :: re, im
  integer :: i, mu, c

  do c = 1, 3
     do mu = 1, 4
        tmp = ZERO
        !$omp parallel do reduction(+: tmp)
        do i = 1, volh
           tmp = tmp + in_e(mu,c,i)
        enddo
        !$omp parallel do reduction(+: tmp)
        do i = 1, volh
           tmp = tmp + in_o(mu,c,i)
        enddo
        re = global_sum(dble(tmp))
        im = global_sum(dimag(tmp))
        sc(mu,c) = cmplx( re, im, kind = RKIND)
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine write_vector12_0000_sum(in_e,in_o,name)

  use module_function_decl
  use module_vol
  implicit none
  SPINCOL_FIELD, intent(in) :: in_e, in_o
  character( len = 6), intent(in) :: name

  COMPLEX, dimension(DIM,NCOL) :: sc
  character( len = 10) :: name1, name2

  call global_sum_vector12(sc, in_e, in_o)

  write(name1,10)name,"0000"
  write(name2,10)name,"_sum"

  call write_vector12(0, in_e(1,1,1), name1)
  call write_vector12(0, sc, name2)

10 format (2a)

end

!-------------------------------------------------------------------------------
subroutine write_vector12(pe,sc,name)

  use module_function_decl
  implicit none
  COMPLEX, dimension(DIM,NCOL), intent(in) :: sc
  integer, intent(in) :: pe
  character( len = 10) :: name
  integer :: mu, c


  if (my_pe() == pe) then
     do c = 1, 3
        do mu = 1, 4
           write(*,100)"%%%",name,& 
           my_pe(),mu,c,dble(sc(mu,c)),dimag(sc(mu,c))
        enddo
     enddo
  endif

100 format (1x, 2a, 3i3, 2ES17.7)

end

!===============================================================================
