!===============================================================================
!
! module_action.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_action_type

  type :: type_plaq
     integer :: aid ! action    id
     integer :: fid ! force     id
     integer :: sid ! smearing  id
     REAL    :: beta
  end type type_plaq

  type :: type_impg
     integer :: aid ! action    id
     integer :: fid ! force     id
     integer :: sid ! smearing  id
     REAL    :: beta1, beta2, beta3
  end type type_impg

  type :: type_mtilde
     integer :: cid ! clover    id
     integer :: did ! dslash    id
     REAL :: rho, kappa, cswkappa, h, b
  end type type_mtilde

  type :: type_fermi
     integer :: aid ! id of action
     integer :: fid ! id of force
     integer :: rid ! id of rational approx
     integer :: mid1 ! mtilde   id denominator (hasenbhsch heavy)
     integer :: mid2 ! mtilde   id numerator   (hasenbhsch light)
     integer :: type ! EO_NORMAL,EO_RATIONAL
                     ! UN_NORMAL,UN_RATIONAL
                     ! DD_NORMAL,DD_RATIONAL
                     ! HH_NORMAL,HH_RATIONAL
     REAL:: factor
     integer :: nth  ! nth rooted 
     logical :: gencoef
  end type type_fermi

  type :: type_logdet
     integer :: aid ! action id
     integer :: fid ! force  id
     integer, dimension(:), pointer :: cid ! clover id
  end type type_logdet

  type :: type_ffd
     integer :: aid ! action id
     integer :: fid ! force  id
     integer :: sid ! smearing  id
     REAL    :: theta
  end type type_ffd

end

!-------------------------------------------------------------------------------
module module_action
  use module_action_type

  !-----------------------------------------------------------------------------
  ! gauge action
  !-----------------------------------------------------------------------------
  character(len = 10), save :: gauge_action

  REAL, save :: c0, c1, c2, c3

  !
  ! for tadpole improvement
  !
  REAL, save :: last_plaq, plaq_ave, g2, u04
  REAL, dimension(:),pointer, save  :: plaq_hist
  integer, dimension(:),pointer, save  :: n_ave_plaq
  integer, save :: last_n_ave_plaq, plaq_start
  character(len = 14), save:: plaq_file

  !-----------------------------------------------------------------------------
  ! fermi action
  !-----------------------------------------------------------------------------
  character(len = 10), save :: fermi_action
  
  !                  10 of all fermionic type
  integer, dimension(10) :: accumulated_num_mpf, num_mpf


  !-----------------------------------------------------------------------------
  ! action infos
  !-----------------------------------------------------------------------------
  integer, save :: n_dslash_id = 0
  integer, save :: n_fermi_id  = 0
  integer, save :: n_clover_id = 0
  integer, save :: n_mtilde_id = 0

  type :: type_action
     type(type_plaq)  :: plaq
     type(type_impg)  :: impg
     type(type_ffd)   :: ffd
     type(type_logdet) :: logdet
     type(type_mtilde), dimension(:), pointer :: mtilde
     type(type_fermi), dimension(:), pointer :: fermi
  end type type_action

  type(type_action), save :: action


  ! even-odd precnditioning notation
#ifdef DOEDEO
  integer, parameter :: se = 1
  integer, parameter :: so = 0
#else
  integer, parameter :: se = 0
  integer, parameter :: so = 1
#endif

end

!===============================================================================
