!===============================================================================
!
! module_function_decl.F90  -  declaration of functions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2003-2006 Hinnerk Stueben 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_function_decl

  ! functions in misc.F90:

  character(len=10), external :: datum
  character(len=12), external :: uhrzeit
  character(len=20), external :: rechner
  logical, external :: f_exist
  character(len=12), external :: itoa
  character(len=24), external :: i8toa
  character(len=24), external :: REALtoa

  ! processes (-> comm/pes*.F90):

  integer, external :: num_pes

#ifdef CRAY
  integer, intrinsic :: my_pe  ! gives the same id as MPI_Rank in MPI_COMM_WORLD
#else
  integer, external :: my_pe
#endif

  ! processes2 (-> index2.F90)

  integer, external :: coord2pe

  ! global reduction (-> comm/reduction_*.F90)

  REAL, external :: dotprod
  REAL, external :: global_sum
  REAL, external :: global_min
  REAL, external :: global_max


  ! sc-field (-> sc.F90)

  REAL, external    :: sc_norm2
  REAL, external    :: sc_dot
  COMPLEX, external :: sc_cdotc

  ! sc-field (-> sc_r4.F90)

  REAL4, external    :: sc_norm2_r4
  REAL4, external    :: sc_dot_r4
  COMPLEX4, external :: sc_cdotc_r4

  ! random numbers (-> ran/.../ran.F90):

  RAN_NAME, external :: ran_name
  REAL, external     :: ran_number
  SEED, external     :: ran_get_seed

  ! identification of D (-> d/DVersion.F90)

  integer, external :: version_of_d

  ! communication method (-> comm/comm_*.F90)

  COMM_METHOD, external :: comm_method


contains

  ! threads:

  integer function my_thread()
    implicit none
#ifdef _OPENMP
    integer, external :: omp_get_thread_num
    my_thread = omp_get_thread_num()
#else
    my_thread = 0
#endif
  end function my_thread

  integer function num_threads()
    implicit none
#ifdef _OPENMP
    integer, external :: omp_get_max_threads
    num_threads = omp_get_max_threads()
#else
    num_threads = 1
#endif
  end function num_threads

end

!===============================================================================
