#/*
#===============================================================================
#
# module_input.h
#
#-------------------------------------------------------------------------------
#
# Copyright (C) 2003-2010 Hinnerk Stueben
#               2007-2010 Yoshifumi Nakamura
#
# This file is part of BQCD -- Berlin Quantum ChromoDynamics program
#
# BQCD is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BQCD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
#
#-------------------------------------------------------------------------------
#*/
INPUT_INPUT(run, integer, 0)
INPUT_INPUT(comment, character(comment_len), "")

INPUT_INPUT(lattice, INPUT_ARRAY_DIM, (/4, 4, 4, 4/))
INPUT_INPUT(processes, INPUT_ARRAY_DIM, (/1, 1, 1, 1/))
INPUT_INPUT(process_mapping, INPUT_ARRAY_DIM, (/1, 2, 3, 4/))
INPUT_INPUT(boundary_conditions_fermions, INPUT_ARRAY_DIM, (/1, 1, 1, -1/))

INPUT_INPUT(ensembles, integer, INPUT_DEFAULT_ENSEMBLES)

INPUT_INPUT(beta, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(kappa, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(csw, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(h, INPUT_ARRAY_ENSEMBLES, "0.0")

INPUT_INPUT(tempering_swap_sequence, character(word_len), "random")
INPUT_INPUT(tempering_steps_without, integer, 0)

INPUT_INPUT(hmc_model, character, "A")
INPUT_INPUT(hmc_trajectory_length, INPUT_ARRAY_ENSEMBLES, "1")
INPUT_INPUT(hmc_steps, INPUT_ARRAY_ENSEMBLES, "0")
INPUT_INPUT(hmc_rho, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(hmc_m_scale, INPUT_ARRAY_ENSEMBLES, "1")
INPUT_INPUT(hmc_m_scale2, INPUT_ARRAY_ENSEMBLES, "1")
INPUT_INPUT(hmc_accept_first, integer, 0)
INPUT_INPUT(hmc_test, integer, 0)

INPUT_INPUT(start_configuration, character(word_len), "cold")
INPUT_INPUT(start_info_file, INPUT_ARRAY_ENSEMBLES, "")
INPUT_INPUT(start_ildg_file, FILENAME, "")
INPUT_INPUT(start_random, character(para_len), "default")

INPUT_INPUT(mc_total_steps, integer, 1)
INPUT_INPUT(mc_steps, integer, 1)
INPUT_INPUT(mc_save_frequency, integer, 0)

INPUT_INPUT(solver_rest, character(para_len), "1e-8")
INPUT_INPUT(solver_maxiter, integer, 100)
INPUT_INPUT(solver_ignore_no_convergence, integer, 0)
INPUT_INPUT(solver_check_solution, integer, 0)
INPUT_INPUT(solver_mre_vectors, integer, 0)

INPUT_INPUT(ran_ranlux_level, integer, 2)

INPUT_INPUT(measure_cooling_list, FILENAME, "")
INPUT_INPUT(measure_polyakov_loop, integer, 0)
INPUT_INPUT(measure_traces, integer, 0)

INPUT_INPUT(tuning_xbound_sc_i, integer, 1)
INPUT_INPUT(tuning_xbound_g_i, integer, 1)
INPUT_INPUT(tuning_cg_precision_r4, integer, 0)
INPUT_INPUT(tuning_cg_loop_blocking, integer, 0)
INPUT_INPUT(tuning_cg_block_length, integer, 256)

INPUT_INPUT(io_restart_format, character(word_len), "bqcd")
INPUT_INPUT(io_conf_format, character(word_len), "bqcd")

INPUT_INPUT(ildg_filename_prefix, FILENAME, "bqcd")
INPUT_INPUT(ildg_filename_extension, FILENAME, "lime")
INPUT_INPUT(ildg_template_ensemble, FILENAME, "")
INPUT_INPUT(ildg_template_conf, FILENAME, "")

INPUT_INPUT(ildg_markov_chain_uri, character(url_len), "mc://UNDEFINED")
INPUT_INPUT(ildg_data_lfn_path, character(url_len), "lfn://UNDEFINED")
INPUT_INPUT(ildg_precision, integer, 64)
INPUT_INPUT(ildg_precision_restart, integer, 64)

INPUT_INPUT(ildg_participant_name, character(string_len), "UNDEFINED")
INPUT_INPUT(ildg_participant_institution, character(string_len), "UNDEFINED")
INPUT_INPUT(ildg_machine_name, character(string_len), "UNDEFINED")
INPUT_INPUT(ildg_machine_institution, character(string_len), "UNDEFINED")
INPUT_INPUT(ildg_machine_type, character(string_len), "UNDEFINED")





INPUT_INPUT(ddlattice, INPUT_ARRAY_DIM, (/1, 1, 1, 1/))

INPUT_INPUT(gauge_action, character(para_len), "WILSON")
INPUT_INPUT(fermi_action, character(para_len), "NON")

INPUT_INPUT(c0, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(c1, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(c2, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(c3, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(u04, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(tadpole_csw, integer, 0)
INPUT_INPUT(fix_coefficients_tadpole, integer, 0)
INPUT_INPUT(fix_coefficients_tadpole_by_u04, integer, 0)

INPUT_INPUT(kappa_strange, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(n_stout, INPUT_ARRAY_ENSEMBLES, "0")
INPUT_INPUT(alpha, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(theta, INPUT_ARRAY_ENSEMBLES, "0.0")

INPUT_INPUT(hmc_integrator1, character(para_len), "LPFSTS")
INPUT_INPUT(hmc_integrator2, character(para_len), "LPFSTS")
INPUT_INPUT(hmc_integrator3, character(para_len), "LPFSTS")
INPUT_INPUT(hmc_integrator4, character(para_len), "NON")
INPUT_INPUT(hmc_integrator5, character(para_len), "NON")
INPUT_INPUT(hmc_integrator6, character(para_len), "NON")

INPUT_INPUT(hmc_rho2, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(hmc_rho3, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(hmc_rho4, INPUT_ARRAY_ENSEMBLES, "0.0")
INPUT_INPUT(hmc_m_scale3, INPUT_ARRAY_ENSEMBLES, "1")
INPUT_INPUT(hmc_m_scale4, INPUT_ARRAY_ENSEMBLES, "1")
INPUT_INPUT(hmc_m_scale5, INPUT_ARRAY_ENSEMBLES, "1")


INPUT_INPUT(hmc_mpf_mass, integer, 0)
INPUT_INPUT(hmc_mpf_rhmc_l, integer, 0)
INPUT_INPUT(hmc_mpf_rhmc_s, integer, 0)
INPUT_INPUT(hmc_mpf_un_mass, integer, 0)
INPUT_INPUT(hmc_mpf_un_rhmc_l, integer, 0)
INPUT_INPUT(hmc_mpf_un_rhmc_s, integer, 0)
INPUT_INPUT(hmc_mpf_dd_mass, integer, 0)
INPUT_INPUT(hmc_mpf_dd_rhmc_l, integer, 0)
INPUT_INPUT(hmc_mpf_dd_rhmc_s, integer, 0)
INPUT_INPUT(hmc_mpf_hh_mass, integer, 0)
INPUT_INPUT(hmc_mpf_hh_rhmc_l, integer, 0)
INPUT_INPUT(hmc_mpf_hh_rhmc_s, integer, 0)

INPUT_INPUT(hmc_hkappa, integer, 0)
INPUT_INPUT(hmc_dsf1_un, integer, 0)
INPUT_INPUT(hmc_dsf2_un, integer, 0)
INPUT_INPUT(hmc_dsf1_dd, integer, 0)
INPUT_INPUT(hmc_dsf1_hh, integer, 0)
INPUT_INPUT(hmc_dsf2_dd, integer, 0)
INPUT_INPUT(hmc_dsf2_hh, integer, 0)

INPUT_INPUT(hmc_dsf_erl, integer, 0)
INPUT_INPUT(hmc_dsf_drl, integer, 0)
INPUT_INPUT(hmc_dsf_hrl, integer, 0)

INPUT_INPUT(hmc_dsf_ers, integer, 0)
INPUT_INPUT(hmc_dsf_drs, integer, 0)
INPUT_INPUT(hmc_dsf_hrs, integer, 0)

INPUT_INPUT(hmc_dsf1_mtmp, integer, 0)
INPUT_INPUT(hmc_dsf2_mtmp, integer, 0)
INPUT_INPUT(hmc_dsf3_mtmp, integer, 0)
INPUT_INPUT(hmc_dsf4_mtmp, integer, 0)
INPUT_INPUT(hmc_dsf5_mtmp, integer, 0)
INPUT_INPUT(hmc_dsd,  integer, 0)
INPUT_INPUT(hmc_dsg,  integer, 0)
INPUT_INPUT(hmc_dsig, integer, 0)
INPUT_INPUT(rhmc_fraction_u1, integer, 40)
INPUT_INPUT(rhmc_fraction_u2, integer, 0)
INPUT_INPUT(rhmc_roughness_level, integer, 0)

INPUT_INPUT(solver_rest_md, character(para_len), "1e-8")
INPUT_INPUT(solver_rest_cg_ritz, character(para_len), "1e-8")
INPUT_INPUT(solver_rest_domain_mc,  character(para_len), "1e-10")
INPUT_INPUT(solver_rest_domain_md,  character(para_len), "1e-10")
INPUT_INPUT(solver_rest_domain_sap, character(para_len), "1e-3")
INPUT_INPUT(solver_rest_domain_def, character(para_len), "1e-10")
INPUT_INPUT(solver_stopping_criterion, integer, 1)
INPUT_INPUT(solver_outer_solver,  character(para_len), "cg")
INPUT_INPUT(solver_inner_solver,  character(para_len), "cg")
INPUT_INPUT(solver_outer_steps, integer, 20)

INPUT_INPUT(solver_defsapgcr_nkv, integer, 16)
INPUT_INPUT(solver_defsapgcr_nsp, integer, 8)
INPUT_INPUT(solver_numkrylov_mtdmtdd, integer, 10)
INPUT_INPUT(solver_numdefvec_mtdmtdd, integer, 0)
INPUT_INPUT(fullsolver,  character(para_len), "eo")

INPUT_INPUT(measure_minmax, integer, 0)
INPUT_INPUT(measure_rhmc_forces, integer, 0)
INPUT_INPUT(measure_correlations, integer, 0)
INPUT_INPUT(measurement_only, integer, 0)
INPUT_INPUT(measurement_start_traj, integer, 0)
INPUT_INPUT(measurement_file_name, character(para_len), "bqcd")

INPUT_INPUT(boundary_sf, integer, 0)
INPUT_INPUT(nonactivelink, integer, 0)
INPUT_INPUT(tuning_approx_range, integer, 0)
INPUT_INPUT(tuning_approx_range_list, FILENAME, "")

!
! TEST
!
INPUT_INPUT(hmc_mpf_hhdd_test, integer, 0)
INPUT_INPUT(hmc_mpf_hhbb_test, integer, 0)
INPUT_INPUT(hmc_dsf_hhdd_test, integer, 0)
INPUT_INPUT(hmc_dsf_hhbb_test, integer, 0)
