!===============================================================================
!
! ranf/ran.F90 - random numbers
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
RAN_NAME function ran_name()

  implicit none

  ran_name = "ranf"
end

!-------------------------------------------------------------------------------
subroutine ran_init(seed)

  implicit none
  SEED, intent(in)  :: seed
  SEED              :: null = 0

  call ranset(seed, null)
end

!-------------------------------------------------------------------------------
REAL function ran_number()

   implicit none

#ifdef CRAY
  real(8), intrinsic :: ranf
#else
  real(8), external :: ranf
#endif

   ran_number = ranf()
end

!-------------------------------------------------------------------------------
subroutine ran_vector(x, n)

   implicit none
   integer, intent(in) :: n
   REAL, intent(out)   :: x(n)
   integer             :: i

#ifdef CRAY
  real(8), intrinsic :: ranf
#else
  real(8), external :: ranf
#endif

   do i = 1, n
      x(i) = ranf()
   enddo
end

!-------------------------------------------------------------------------------
subroutine ran_skip(n)

  implicit none
  integer, intent(in) :: n
  SEED                :: seed, n_skip

  n_skip = n
  call ranget(seed)
  call ranset(seed, n_skip)
end

!-------------------------------------------------------------------------------
subroutine ran_dump()

  use module_function_decl
  implicit none
  FILENAME, external :: ran_file
  SEED               :: seed

  call ran_check()
  call ranget(seed)

  if (my_pe() == 0) then
     open(URAN, file = ran_file(), action = "write")
     write(URAN, *) seed
     close(URAN)
  endif

end

!-------------------------------------------------------------------------------
subroutine ran_restore()

  implicit none
  FILENAME, external :: ran_file
  SEED               :: seed
  SEED               :: null = 0

  open(URAN, file = ran_file(), action = "read", status = "old")
  read(URAN, *) seed
  close(URAN)

  call ranset(seed, null)
end

!-------------------------------------------------------------------------------
SEED function ran_get_seed()

  implicit none

  call ranget(ran_get_seed)
end

!-------------------------------------------------------------------------------
subroutine ran_check()

  use module_function_decl
  implicit none
  SEED :: my_seed, seed

  call ranget(my_seed)

  if (my_pe() == 0) then
     seed = my_seed
  else 
     seed = -1
  endif

  call comm_broadcast_i8(seed, 1)

  ASSERT(seed == my_seed)
end

!===============================================================================
