!===============================================================================
!
! ran/ranlux-3.2/ran.F90 - luxury (Luscher) random numbers
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_ranlux

  integer, parameter :: state_size = 105
  integer, external  :: rlxd_size_f
  integer            :: state(state_size)
  integer            :: my_state(state_size)

  character          :: c_level(2) = (/'1', '2'/)
end

!-------------------------------------------------------------------------------
RAN_NAME function ran_name()

  use module_input
  implicit none
  character :: c_level(2) = (/'1', '2'/)
  integer   :: level

  level = input%ran_ranlux_level

  CHECK(level == 1 .or. level == 2)

  ran_name = "ranlux-3.2 level " // c_level(level)
end

!-------------------------------------------------------------------------------
subroutine ran_init(seed)

  use module_input
  use module_ranlux
  implicit none
  SEED, intent(in) :: seed
!!SEED, parameter  :: mask31 = 2**31 - 1
  SEED, parameter  :: mask31 = 2147483647
  integer          :: seed32
  integer          :: level

  level = input%ran_ranlux_level

  CHECK(level == 1 .or. level == 2)

  if (seed > mask31) call warn("ran_init(): seed cut to lowest 31 bits")

  seed32 = iand(mask31, seed)

  call rlxd_init_f(level, seed32)
end

!-------------------------------------------------------------------------------
REAL function ran_number()

   implicit none

   call ranlxd_f(ran_number, 1)
end

!-------------------------------------------------------------------------------
subroutine ran_vector(x, n)

   implicit none
   integer, intent(in) :: n
   REAL, intent(out)   :: x(n)
   integer             :: i

   call ranlxd_f(x, n)
end

!-------------------------------------------------------------------------------
subroutine ran_skip(n)

  implicit none
  integer, intent(in) :: n
  REAL                :: x
  integer             :: i

  do i = 1, n
     call ranlxd_f(x, 1)
  enddo
end

!-------------------------------------------------------------------------------
subroutine ran_dump()

  use module_function_decl
  use module_ranlux
  implicit none
  FILENAME, external :: ran_file
  integer            :: i

  ASSERT(state_size == rlxd_size_f())

  call ran_check()

  call rlxd_get_f(state)

  if (my_pe() == 0) then
     open(URAN, file = ran_file(), action = "write")
     do i = 1, state_size
        write(URAN, *) state(i)
     enddo
     close(URAN)
  endif

end

!-------------------------------------------------------------------------------
subroutine ran_restore()

  use module_ranlux
  implicit none
  FILENAME, external :: ran_file
  integer            :: i

  ASSERT(state_size == rlxd_size_f())

  open(URAN, file = ran_file(), action = "read", status = "old")
  do i = 1, state_size
     read(URAN, *) state(i)
  enddo
  close(URAN)

  call rlxd_reset_f(state)
end

!-------------------------------------------------------------------------------
SEED function ran_get_seed()

  implicit none

  ran_get_seed = -1
end

!-------------------------------------------------------------------------------
subroutine ran_check()

  use module_function_decl
  use module_ranlux
  implicit none
  integer :: i

  call rlxd_get_f(my_state)

  if (my_pe() == 0) then
     state(:) = my_state(:)
  else 
     state(:) = -1
  endif

  call comm_broadcast_i4(state, state_size)

  do i = 1, state_size
     ASSERT(state(i) == my_state(i))
  enddo
end

!===============================================================================
