!===============================================================================
!
! ran/ranlux-3.2/ran_test.F90 - test of ran.F90 for ranlux
!
!------------------------------------------------------------------------------
!
! Copyright (C) 2008 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!------------------------------------------------------------------------------
# include "defs.h"

!------------------------------------------------------------------------------
program ran_test

  use module_input
  implicit none
  FILENAME, external :: ran_file
  RAN_NAME, external :: ran_name
  REAL, external     :: ran_number
  REAL               :: xdn(0:47)
  REAL               :: x(0:47)
  REAL               :: r
  REAL               :: base = 2.0_8**48
  SEED               :: seed = 32767
  integer            :: i, level

  input%ran_ranlux_level = 1

  xdn(0) = 135665102723086.0_8
  xdn(1) = 259840970195871.0_8
  xdn(2) = 110726726657103.0_8
  xdn(3) = 53972500363809.0_8
  xdn(4) = 199301297412157.0_8
  xdn(5) = 63744794353870.0_8
  xdn(6) = 178745978725904.0_8
  xdn(7) = 243549380863176.0_8
  xdn(8) = 244796821836177.0_8
  xdn(9) = 223788809121855.0_8
  xdn(10) = 113720856430443.0_8
  xdn(11) = 124607822268499.0_8
  xdn(12) = 25705458431399.0_8
  xdn(13) = 155476863764950.0_8
  xdn(14) = 195602097736933.0_8
  xdn(15) = 183038707238950.0_8
  xdn(16) = 62268883953527.0_8
  xdn(17) = 157047615112119.0_8
  xdn(18) = 58134973897037.0_8
  xdn(19) = 26908869337679.0_8
  xdn(20) = 259927185454290.0_8
  xdn(21) = 130534606773507.0_8
  xdn(22) = 205295065526788.0_8
  xdn(23) = 40201323262686.0_8
  xdn(24) = 193822255723177.0_8
  xdn(25) = 239720285097881.0_8
  xdn(26) = 54433631586673.0_8
  xdn(27) = 31313178820772.0_8
  xdn(28) = 152904879618865.0_8
  xdn(29) = 256187025780734.0_8
  xdn(30) = 110292144635528.0_8
  xdn(31) = 26555117184469.0_8
  xdn(32) = 228913371644996.0_8
  xdn(33) = 126837665590799.0_8
  xdn(34) = 141069100232139.0_8
  xdn(35) = 96171028602910.0_8
  xdn(36) = 259271018918511.0_8
  xdn(37) = 65257892816619.0_8
  xdn(38) = 14254344610711.0_8
  xdn(39) = 137794868158301.0_8
  xdn(40) = 269703238916504.0_8
  xdn(41) = 35782602710520.0_8
  xdn(42) = 51447305327263.0_8
  xdn(43) = 247852246697199.0_8
  xdn(44) = 65072958134912.0_8
  xdn(45) = 273325640150591.0_8
  xdn(46) = 2768714666444.0_8
  xdn(47) = 173907458721736.0_8
  

  write(6,*) "ran_test: testing ", trim(ran_name())

  write(6,*) "ran_test: ran_skip() and ran_number()"
  call ran_init(seed)
  call ran_skip(19 * 99 + 39)
  do i = 0, 47
     ASSERT(ran_number() * base == xdn(i))
  enddo

  write(6,*) "ran_test: ran_vector()"
  call ran_init(seed)
  call ran_skip(19 * 99 + 39)
  call ran_vector(x, 48)
  do i = 0, 47
     ASSERT(x(i) * base == xdn(i))
  enddo

  write(6,*) "ran_test: ran_dump() and ran_restore() [dump: ", &
             trim(ran_file()), "]"
  call ran_init(seed)
  call ran_skip(11 * 99)
  call ran_dump()
  call ran_init(seed)
  call ran_restore()
  call ran_skip(8 * 99 + 39)
  do i = 0, 47
     ASSERT(ran_number() * base == xdn(i))
  enddo

  write(6,*) "ran_test: all tests passed"
end

!------------------------------------------------------------------------------
FILENAME function ran_file()

  implicit none

  ran_file = "ran_test.dump"
end

!==============================================================================
