!===============================================================================
!
! copy
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Hinnerk Stueben
!               2009 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

# define GAUGE_FIELD_R4 complex(4), dimension (NCOL, NCOL, volh_tot, EVEN:ODD, DIM)
# define CLOVER_FIELD_A_R4 type(type_clover_a_r4), dimension(2, volh, EVEN:ODD)
# define CLOVER_FIELD_B_R4 type(type_clover_b_r4), dimension(2, volh, EVEN:ODD)

!-------------------------------------------------------------------------------
subroutine gauge_copy_r8_to_r4(u4, u8)

  use module_vol
  implicit none
  GAUGE_FIELD_R4, intent(out) :: u4
  GAUGE_FIELD   , intent(in)  :: u8
  integer        :: i, mu, eo, c1, c2

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do private(c1, c2)
        do i = 1, volh_tot
           do c2 = 1, NCOL
              do c1 = 1, NCOL
                 u4(c1, c2, i, eo, mu) = u8(c1, c2, i, eo, mu)
              enddo
           enddo
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine clover_a_copy_r8_to_r4(a4, a8)
  use typedef_clover
  use typedef_clover_r4
  use module_vol
  implicit none
  CLOVER_FIELD_A_R4, intent(out) :: a4
  CLOVER_FIELD_A   , intent(in)  :: a8
  integer        :: i, j, eo

  do eo = EVEN, ODD
     !$omp parallel do private(j)
     do i = 1, VOLH
        do j = 1, 2
           call type_clover_a_copy_r8_to_r4(a4(j, i, eo), a8(j, i, eo))
        enddo
     enddo
  enddo


end

!-------------------------------------------------------------------------------
subroutine clover_b_copy_r8_to_r4(b4, b8)
  use typedef_clover
  use typedef_clover_r4
  use module_vol
  implicit none
  CLOVER_FIELD_B_R4, intent(out) :: b4
  CLOVER_FIELD_B   , intent(in)  :: b8
  integer        :: i, j, eo

  do eo = EVEN, ODD
     !$omp parallel do private(j)
     do i = 1, VOLH
        do j = 1, 2
           call type_clover_b_copy_r8_to_r4(b4(j, i, eo), b8(j, i, eo))
        enddo
     enddo
  enddo
  
end

!-------------------------------------------------------------------------------
subroutine type_clover_a_copy_r8_to_r4(a4, a8)
  use typedef_clover
  use typedef_clover_r4
  implicit none
  type(type_clover_a_r4), intent(out) :: a4
  type(type_clover_a),    intent(in)  :: a8

  a4%i11 = a8%i11
  a4%i22 = a8%i22
  a4%i12 = a8%i12
  a4%i13 = a8%i13
  a4%i14 = a8%i14
  a4%i15 = a8%i15
  a4%i16 = a8%i16
  a4%i23 = a8%i23
  a4%i24 = a8%i24
  a4%i25 = a8%i25
  a4%i26 = a8%i26
  a4%i33 = a8%i33
  a4%i44 = a8%i44
  a4%i34 = a8%i34
  a4%i35 = a8%i35
  a4%i36 = a8%i36
  a4%i45 = a8%i45
  a4%i46 = a8%i46
  a4%i55 = a8%i55
  a4%i66 = a8%i66
  a4%i56 = a8%i56

end

!-------------------------------------------------------------------------------
subroutine type_clover_b_copy_r8_to_r4(b4, b8)
  use typedef_clover
  use typedef_clover_r4
  implicit none
  type(type_clover_b_r4), intent(out) :: b4
  type(type_clover_b),    intent(in)  :: b8

  b4%i21 = b8%i21
  b4%i31 = b8%i31
  b4%i32 = b8%i32
  b4%i41 = b8%i41
  b4%i42 = b8%i42
  b4%i43 = b8%i43
  b4%i51 = b8%i51
  b4%i52 = b8%i52
  b4%i53 = b8%i53
  b4%i54 = b8%i54
  b4%i61 = b8%i61
  b4%i62 = b8%i62
  b4%i63 = b8%i63
  b4%i64 = b8%i64
  b4%i65 = b8%i65
  b4%i11 = b8%i11
  b4%i22 = b8%i22
  b4%i33 = b8%i33
  b4%i44 = b8%i44
  b4%i55 = b8%i55
  b4%i66 = b8%i66

end

!-------------------------------------------------------------------------------
subroutine type_clover_a_copy_r8(out, in)
  use typedef_clover
  implicit none
  type(type_clover_a),    intent(out) :: out
  type(type_clover_a),    intent(in)  :: in

  out%i11 = in%i11
  out%i22 = in%i22
  out%i12 = in%i12
  out%i13 = in%i13
  out%i14 = in%i14
  out%i15 = in%i15
  out%i16 = in%i16
  out%i23 = in%i23
  out%i24 = in%i24
  out%i25 = in%i25
  out%i26 = in%i26
  out%i33 = in%i33
  out%i44 = in%i44
  out%i34 = in%i34
  out%i35 = in%i35
  out%i36 = in%i36
  out%i45 = in%i45
  out%i46 = in%i46
  out%i55 = in%i55
  out%i66 = in%i66
  out%i56 = in%i56

end

!-------------------------------------------------------------------------------
subroutine type_clover_b_copy_r8(out, in)
  use typedef_clover
  implicit none
  type(type_clover_b),    intent(out) :: out
  type(type_clover_b),    intent(in)  :: in

  out%i21 = in%i21
  out%i31 = in%i31
  out%i32 = in%i32
  out%i41 = in%i41
  out%i42 = in%i42
  out%i43 = in%i43
  out%i51 = in%i51
  out%i52 = in%i52
  out%i53 = in%i53
  out%i54 = in%i54
  out%i61 = in%i61
  out%i62 = in%i62
  out%i63 = in%i63
  out%i64 = in%i64
  out%i65 = in%i65
  out%i11 = in%i11
  out%i22 = in%i22
  out%i33 = in%i33
  out%i44 = in%i44
  out%i55 = in%i55
  out%i66 = in%i66

end

!-------------------------------------------------------------------------------
!!subroutine gauge_copy(out, in)
!!  use module_vol
!!  implicit none
!!  P_GAUGE_FIELD :: out, in
!!  integer :: mu, eo, i
!!
!!  do mu = 1, DIM
!!     do eo = EVEN, ODD
!!        !$omp parallel do
!!        do i = 1, volh
!!           out(1, 1, i, eo, mu) = in(1, 1, i, eo, mu)
!!           out(2, 1, i, eo, mu) = in(2, 1, i, eo, mu)
!!           out(3, 1, i, eo, mu) = in(3, 1, i, eo, mu)
!!           out(1, 2, i, eo, mu) = in(1, 2, i, eo, mu)
!!           out(2, 2, i, eo, mu) = in(2, 2, i, eo, mu)
!!           out(3, 2, i, eo, mu) = in(3, 2, i, eo, mu)
!!           out(1, 3, i, eo, mu) = in(1, 3, i, eo, mu)
!!           out(2, 3, i, eo, mu) = in(2, 3, i, eo, mu)
!!           out(3, 3, i, eo, mu) = in(3, 3, i, eo, mu)
!!        enddo
!!     enddo
!!  enddo
!!
!!end

!-------------------------------------------------------------------------------
subroutine copy_r8_from_r4(out, in, n)

  implicit none
  integer, intent(in) :: n
  REAL8, intent(out)  :: out(n)
  REAL4, intent(in)   :: in(n)
  integer             :: i

  !$omp parallel do
  do i = 1, n
     out(i) = in(i)
  enddo
end

!-------------------------------------------------------------------------------
subroutine copy_r4_from_r8(out, in, n)

  implicit none
  integer, intent(in) :: n
  REAL4, intent(out)  :: out(n)
  REAL8, intent(in)   :: in(n)
  integer             :: i

  !$omp parallel do
  do i = 1, n
     out(i) = in(i)
  enddo
end

!===============================================================================
