!===============================================================================
!
! init_confs2.F90 - operations on gauge field and 
!                   pseudo fermion field configurations
!                      -- rhmc
!                      -- multiple pseudofermions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2006-2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_confs2(para, conf)
  use module_stout_field
  use module_field
  use module_action
  use typedef_para
  use module_p_interface
  use module_p_interface_r4
  use module_switches
  use module_dd
  implicit none

  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf
  integer                               :: i, count

  DEBUG2S("Start: init_confs")

  do i = 1, para%n_temper
     call allocate_g_field(conf(i)%u)
     if (para%start == START_HOT .or. para%start == START_COLD) then
        call init_u(conf(i)%u, para%start)
        conf(i)%former = i
     endif
  enddo

  if (para%start == START_CONT) call conf_read(.true., para, conf)
  if (para%start == START_FILE) call conf_read(.false., para, conf)

  do i = 1, para%n_temper
     if (switches%boundary_sf) call schr_boundary_gauge(conf(i)%u)
  enddo

!--------------------------------------------------------------------------
  count = 1
  if (switches%dynamical) then
     count = count + 1
     if (switches%clover) then
        count = count + 1
        call init_clover_field(para%hmc(1))
     endif
     call init_fermi_field(para%hmc(1))
     call init_usc_dd(ld%lat%volh, ld%lat%volh_tot, ld%lat%num_ip)
  else
     nullify(mpf)
     nullify(action%fermi)
     nullify(action%mtilde)
  endif
  allocate(gauge(count)) ! for u ds cl
  allocate(gauge_r4(count)) ! for u ds cl
  gauge(1:count)%init = .true.
  do i = 2, count
     nullify(gauge(i)%u)
     call allocate_g_field(gauge(i)%u)
  enddo
  if (switches%dynamical) then
     nullify(gauge_r4(2)%u)
     call allocate_g_field_r4(gauge_r4(2)%u)
     if (switches%stout_ds) then
        gauge(2)%sid = 1
     else
        gauge(2)%sid = 0
     endif
!!#ifdef LIBDI
!!     call init_d_g_u_reorder()  -> init_d
!!#endif
#ifdef BAGEL
     call allocate_bagel_d()
#endif
  endif
  if (switches%clover) then
     if (switches%stout_cl) then
        gauge(3)%sid = 1
     else
        gauge(3)%sid = 0
     endif
  endif
  gauge(1)%u => conf(1)%u    !!! conf(1)%u must have "target" attribution

  if (switches%stout) then
     allocate(stout(1))
     stout(1)%n = para%hmc(1)%n_stout
     stout(1)%alpha = para%hmc(1)%alpha
     nullify(stout(1)%iter)
     stout(1:size(stout))%init = .true.
  endif

  call read_plaq(para)

  DEBUG2S("End: init_confs")

end

!-------------------------------------------------------------------------------
subroutine init_clover_field(para)
  use module_action
  use module_field
  use typedef_para
  use module_input
  use module_p_interface
  use module_p_interface_r4
  implicit none
  type(hmc_para) :: para
  integer, dimension(:), allocatable :: nf
  REAL   , dimension(:), allocatable :: theta, csw, kappa
  integer         :: i, n_logdet, cid
  integer, external :: clover_id

  !
  ! here count number of logdet and clover fields
  !

  n_logdet = 1

  if (para%kappa_strange /= ZERO) then
     if (para%kappa_strange /= para%kappa) then
        n_logdet = n_logdet + 1
     else
        if (para%theta /= ZERO ) then
           n_logdet = n_logdet + 1
        endif
     endif
  endif
  cid = n_logdet
  if (input%hmc_hkappa /= 0) then
     if ( input%hmc_mpf_mass >1 ) cid = cid + input%hmc_mpf_mass - 1
     if ( input%hmc_mpf_un_mass >1 ) cid = cid + input%hmc_mpf_un_mass - 1
     if ( input%hmc_mpf_dd_mass >1 ) cid = cid + input%hmc_mpf_dd_mass - 1
!!     if ( input%hmc_mpf_hh_mass >1 ) cid = cid + input%hmc_mpf_hh_mass - 1
  endif

  !
  ! here allocate clover fields clover and clover_r4
  ! also temporally nf, theta, csw 
  !
  if (.not. associated(clover)) then
     nullify(clover)
     nullify(clover_r4)
     allocate(clover(cid)) ! # of kappa
     allocate(clover_r4(cid)) ! # of kappa
     allocate(nf(cid))
     allocate(theta(cid))
     allocate(csw(cid))
     allocate(kappa(cid))
     do i = 1, cid
        nullify(clover(i)%a)
        nullify(clover(i)%i)
        nullify(clover(i)%b)
        call allocate_clover_field_a(clover(i)%a)
        call allocate_clover_field_a(clover(i)%i)
        call allocate_clover_field_b(clover(i)%b)
        nullify(clover_r4(i)%a)
        nullify(clover_r4(i)%i)
        nullify(clover_r4(i)%b)
        call allocate_clover_field_a_r4(clover_r4(i)%a)
        call allocate_clover_field_a_r4(clover_r4(i)%i)
        call allocate_clover_field_b_r4(clover_r4(i)%b)
     enddo
  endif

  cid = 1
  nf(cid) = 2
  theta(cid) = para%theta
  csw(cid) = para%csw
  kappa(cid) = para%kappa

  if (para%kappa_strange /= ZERO) then
     if (para%kappa_strange == para%kappa .and. para%theta == ZERO ) then
        nf(1) = nf(1) + 1
     else
        cid = cid + 1
        nf(cid) = 1
        theta(cid) = ZERO
        csw(cid) = para%csw
        kappa(cid) = para%kappa_strange
     endif
  endif

  allocate(action%logdet%cid(n_logdet))
  action%logdet%aid       = 3
  action%logdet%fid       = 3

  do i = 1, n_logdet
     action%logdet%cid(i)  = clover_id(kappa(i), csw(i), theta(i), nf(i), .true.)
  enddo

end

!-------------------------------------------------------------------------------
subroutine init_fermi_field(para)
  use module_action
  use module_field
  use typedef_para
  use module_input
  use module_p_interface
  use module_dd
  use module_vol
  implicit none
  type(hmc_para) :: para
  integer         :: i, num_mpfs, defnormal(4), i_nm

  DEBUG2S("Start: init_fermi_field")

  num_mpf( EO_NORMAL   ) = input%hmc_mpf_mass
  num_mpf( UN_NORMAL   ) = input%hmc_mpf_un_mass
  num_mpf( DD_NORMAL   ) = input%hmc_mpf_dd_mass
  num_mpf( HH_NORMAL   ) = input%hmc_mpf_hh_mass

  num_mpf( EO_RATIONAL ) = input%hmc_mpf_rhmc_l + input%hmc_mpf_rhmc_s
  num_mpf( UN_RATIONAL ) = input%hmc_mpf_un_rhmc_l + input%hmc_mpf_un_rhmc_s
  num_mpf( DD_RATIONAL ) = input%hmc_mpf_dd_rhmc_l + input%hmc_mpf_dd_rhmc_s
  num_mpf( HH_RATIONAL ) = input%hmc_mpf_hh_rhmc_l + input%hmc_mpf_hh_rhmc_s

  num_mpf( HH_NORMAL_DD ) = input%hmc_mpf_hhdd_test
  num_mpf( HH_NORMAL_BB ) = input%hmc_mpf_hhbb_test

  defnormal(1) = EO_NORMAL
  defnormal(2) = UN_NORMAL
  defnormal(3) = DD_NORMAL
  defnormal(4) = HH_NORMAL


  num_mpfs =0
  do i = 1,10
     accumulated_num_mpf(i)=num_mpfs
     num_mpfs = num_mpfs + num_mpf(i)
  enddo

  if (.not. associated(action%fermi)) then
     nullify(action%fermi)
     nullify(action%mtilde)
     allocate(action%fermi( num_mpfs) )
     allocate(action%mtilde(num_mpfs) )
  endif

  !
  ! hmc_mpf_mass
  !
do i_nm = 1, 4
  if (input%hmc_hkappa /= 0) then
     if (num_mpf( defnormal(i_nm) ) >= 1 ) then
     if (num_mpf( defnormal(i_nm) ) == 1 ) then
        call fermi_id(para%kappa,ZERO    ,para%csw, ZERO    ,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%rho  ,ZERO    ,para%csw, ZERO    ,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 2 ) then
     if (num_mpf( defnormal(i_nm) ) == 2 ) then
        call fermi_id(para%kappa,para%rho,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%rho2, para%rho,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 3 ) then
     if (num_mpf( defnormal(i_nm) ) == 3 ) then
        call fermi_id(para%kappa,para%rho2,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%rho3, para%rho2,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 4 ) then
     if (num_mpf( defnormal(i_nm) ) == 4 ) then
        call fermi_id(para%kappa,para%rho3,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%rho4, para%rho3,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) == 5 ) then
        call fermi_id(para%kappa,para%rho4,para%csw, para%csw,ZERO,ZERO,para%theta,2, defnormal(i_nm))
     endif
  else
     if (num_mpf( defnormal(i_nm) ) >= 1 ) then
     if (num_mpf( defnormal(i_nm) ) == 1 ) then
        call fermi_id(para%kappa,ZERO      ,para%csw,ZERO    ,ZERO     ,ZERO     ,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%kappa,ZERO      ,para%csw,ZERO    ,para%rho ,ZERO     ,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 2 ) then
     if (num_mpf( defnormal(i_nm) ) == 2 ) then
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,ZERO     ,para%rho ,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,para%rho2,para%rho ,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 3 ) then
     if (num_mpf( defnormal(i_nm) ) == 3 ) then
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,ZERO     ,para%rho2,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,para%rho3,para%rho2,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) >= 4 ) then
     if (num_mpf( defnormal(i_nm) ) == 4 ) then
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,ZERO     ,para%rho3,para%theta,2, defnormal(i_nm))
     else
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,para%rho4,para%rho3,para%theta,2, defnormal(i_nm))
     endif
     endif
     if (num_mpf( defnormal(i_nm) ) == 5 ) then
        call fermi_id(para%kappa,para%kappa,para%csw,para%csw,ZERO,     para%rho4,para%theta,2, defnormal(i_nm))
     endif
  endif
enddo

  !
  ! hmc_mpf_rhmc_l
  !
  do i = 1, input%hmc_mpf_rhmc_l
     call fermi_id(para%kappa,       ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,2, EO_RATIONAL)
  enddo
  !
  ! hmc_mpf_rhmc_s
  !
  do i = 1, input%hmc_mpf_rhmc_s
    call fermi_id(para%kappa_strange,ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,1, EO_RATIONAL)
  enddo


  !
  ! hmc_mpf_un_rhmc_l
  !
  do i = 1, input%hmc_mpf_un_rhmc_l
     call fermi_id(para%kappa,       ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,2, UN_RATIONAL)
  enddo
  !
  ! hmc_mpf_un_rhmc_s
  !
  do i = 1, input%hmc_mpf_un_rhmc_s
    call fermi_id(para%kappa_strange,ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,1, UN_RATIONAL)
  enddo



  !
  ! hmc_mpf_dd_rhmc_l
  !
  do i = 1, input%hmc_mpf_dd_rhmc_l
     call fermi_id(para%kappa,       ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,2, DD_RATIONAL)
  enddo
  !
  ! hmc_mpf_dd_rhmc_s
  !
  do i = 1, input%hmc_mpf_dd_rhmc_s
    call fermi_id(para%kappa_strange,ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,1, DD_RATIONAL)
  enddo


  !
  ! hmc_mpf_hh_rhmc_l
  !
  do i = 1, input%hmc_mpf_hh_rhmc_l
     call fermi_id(para%kappa,       ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,2, HH_RATIONAL)
  enddo
  !
  ! hmc_mpf_hh_rhmc_s
  !
  do i = 1, input%hmc_mpf_hh_rhmc_s
    call fermi_id(para%kappa_strange,ZERO,para%csw,ZERO,ZERO,ZERO,ZERO,1, HH_RATIONAL)
  enddo





  !!
  !! TEST
  !!
  if (input%hmc_mpf_hhdd_test >= 1 ) then
     call fermi_id(para%kappa,ZERO      ,para%csw,ZERO    ,ZERO ,ZERO     ,para%theta,2, HH_NORMAL_DD)
  endif
  if (input%hmc_mpf_hhbb_test >= 1 ) then
     call fermi_id(para%kappa,ZERO      ,para%csw,ZERO    ,ZERO ,ZERO     ,para%theta,2, HH_NORMAL_BB)
  endif



  if (.not. associated(mpf)) then
  allocate(         mpf( num_mpfs ) )
  do i = 1, num_mpfs
     select case(action%fermi(i)%type)
     case(EO_NORMAL)
        nullify(mpf(i)%sc)
!!        call allocate_sc_field(mpf(i)%sc)
        allocate(mpf(i)%sc(4,3,volh))
     case(DD_NORMAL)
        nullify(mpf(i)%sc)
        call allocate_sc_field(mpf(i)%sc)
     case(EO_RATIONAL)
        nullify(mpf(i)%sc)
!!        call allocate_sc_field(mpf(i)%sc)
        allocate(mpf(i)%sc(4,3,volh))
     case(DD_RATIONAL)
        nullify(mpf(i)%sc)
        call allocate_sc_field(mpf(i)%sc)
     case(HH_NORMAL)
        nullify(mpf(i)%sc)
        call allocate_sc_field(mpf(i)%sc)
!!        allocate(mpf(i)%ddsc(4,3,ld%lat%volh,0:ld%lat%num_ip-1,0:1))
     case(HH_RATIONAL)
        nullify(mpf(i)%sc)
        call allocate_sc_field(mpf(i)%sc)
!!        allocate(mpf(i)%ddsc(4,3,ld%lat%volh,0:ld%lat%num_ip-1,0:1))
     case(UN_NORMAL)
        nullify(mpf(i)%sc_e)
        nullify(mpf(i)%sc_o)
        call allocate_sc_field(mpf(i)%sc_e)
        call allocate_sc_field(mpf(i)%sc_o)
     case(UN_RATIONAL)
        nullify(mpf(i)%sc_e)
        nullify(mpf(i)%sc_o)
        call allocate_sc_field(mpf(i)%sc_e)
        call allocate_sc_field(mpf(i)%sc_o)
     !!
     !! TEST
     !!
     case(HH_NORMAL_DD)
        nullify(mpf(i)%ddsc)
        allocate(mpf(i)%ddsc(4,3,ld%lat%volh,0:ld%lat%num_ip-1,0:1))
     case(HH_NORMAL_BB)
        nullify(mpf(i)%ddsc)
        allocate(mpf(i)%ddsc(4,3,ld%lat%volh,0:ld%lat%num_ip-1,0:1))
     case default
        write(0,*)&
        "init_fermi_field: unknown fermi type:", &
        action%fermi(i)%type
     call die("init_fermi_field: unknown fermi type: ")
     end select
  enddo
  endif

  DEBUG2S("End: init_fermi_field")
end

!-------------------------------------------------------------------------------
integer function mtilde_id(kappa, csw, rho, theta, nf)   !update   n_mtilde_id
  use module_action
  implicit none
  REAL,intent(in) :: rho, kappa, csw, theta 
  integer,intent(in) :: nf
  integer, external :: clover_id, dslash_id
  integer :: i, tmp_cid, tmp_did

  if (kappa == ZERO .and. csw == ZERO) then
     mtilde_id = 0
     return
  endif

  DEBUG2S("Start: mtilde_id")

  tmp_cid = clover_id(kappa, csw, theta, nf, .false.)
  tmp_did = dslash_id(1)

  DEBUG2I("mtilde_id: cid",tmp_cid)

  do i = 1, n_mtilde_id
     if ( action%mtilde(i)%cid == tmp_cid .and. &
          action%mtilde(i)%did == tmp_did .and. &
          abs(action%mtilde(i)%rho - rho)< 1e-10     .and. &
          abs(action%mtilde(i)%kappa -kappa) <1e-10  .and. &
          abs(action%mtilde(i)%cswkappa - csw * kappa)<1e-10   ) then
        mtilde_id = i
        DEBUG2S("End: mtilde_id")
        return
     endif
  enddo

  n_mtilde_id = n_mtilde_id + 1

!write(0,*)"mid:",n_mtilde_id,tmp_cid  ,(action%mtilde(1)%cid == tmp_cid)
!write(0,*)"mid:",n_mtilde_id,tmp_did  ,(action%mtilde(1)%did == tmp_did)
!write(0,*)"mid:",n_mtilde_id,rho      ,(action%mtilde(1)%rho == rho  )
!write(0,*)"mid:",n_mtilde_id,kappa    ,(action%mtilde(1)%kappa == kappa)
!write(0,*)"mid:",n_mtilde_id,csw*kappa,(action%mtilde(1)%cswkappa  == csw * kappa )

  if(n_mtilde_id > size(action%mtilde)) then
     write(0,*) "n_mtilde_id > size(action%mtilde) :: size is",size(action%mtilde)
     stop
  endif

  action%mtilde(n_mtilde_id)%cid   = tmp_cid
  action%mtilde(n_mtilde_id)%did   = tmp_did
  action%mtilde(n_mtilde_id)%rho   = rho
  action%mtilde(n_mtilde_id)%kappa = kappa
  action%mtilde(n_mtilde_id)%cswkappa   = csw * kappa
  action%mtilde(n_mtilde_id)%h     = ZERO
  action%mtilde(n_mtilde_id)%b     = - kappa**2 / (ONE + action%mtilde(n_mtilde_id)%h**2)
  mtilde_id = n_mtilde_id

  DEBUG2S("End: mtilde_id")
end

!-------------------------------------------------------------------------------
integer function clover_id(kappa, csw, theta, nf, care_nf)   !update n_clover_id
  use module_field
  use module_action
  implicit none
  logical,intent(in) :: care_nf
  integer,intent(in) :: nf
  REAL,intent(in) :: kappa, csw, theta
  integer :: i

  if (csw == ZERO) then
     clover_id = 0
     return
  endif

  do i = 1, n_clover_id
     if ( clover(i)%csw   == csw            .and. &
          clover(i)%kappa == kappa          .and. &
          clover(i)%theta == theta/dble(nf) ) then
        if (care_nf) then
           if ( clover(i)%nf == nf )then
              clover_id = i
              return
           endif
        else
           clover_id = i
           return
        endif
     endif
  enddo

  n_clover_id = n_clover_id + 1

  if (n_clover_id > size(clover)) then
     write(0,*)"n_clover_id > size(clover) :: size is", size(clover)
     stop
  endif

  clover(n_clover_id)%csw   = csw
  clover(n_clover_id)%kappa = kappa
  if (kappa == ZERO) then
     clover(n_clover_id)%cswkappa = csw
  else
     clover(n_clover_id)%cswkappa = csw * kappa
  endif
  clover(n_clover_id)%theta = theta / dble(nf)
  clover(n_clover_id)%nf    = nf
  clover(n_clover_id)%init  = .true.
  clover_id = n_clover_id

end

!-------------------------------------------------------------------------------
integer function dslash_id(n)   !update n_dslash_id
  use module_action
  implicit none
  integer,intent(in) :: n

  n_dslash_id = 1
  dslash_id = 1

end

!!!-------------------------------------------------------------------------------
!!integer function action_id(c)   !update n_dslash_id
!!  use module_action
!!  implicit none
!!  character(len=*),intent(in) :: c
!!
!!action_id = 4
!!
!!end
!-------------------------------------------------------------------------------
integer function force_id(type, mid)   !update n_dslash_id
  use module_action
  implicit none
  integer, save :: fid = 3, lasttype=-1, lastmid=-1
  integer, intent(in) :: type, mid

  if (lasttype /= type ) then
     fid = fid + 1
  elseif (type <= 4 .or. (lastmid/=mid)) then
     fid = fid + 1
  endif
  lasttype = type
  lastmid  = mid
  force_id = fid
end

!-------------------------------------------------------------------------------
subroutine fermi_id(kappa1, kappa2, csw1, csw2, rho1, rho2, theta, nf, type) ! update n_fermi_id
  use module_action
  use module_input
  implicit none
  integer, intent(in) :: nf, type
  REAL, intent(in) :: kappa1, csw1, rho1
  REAL, intent(in) :: kappa2, csw2, rho2, theta
  logical, external :: type_rational
  integer, external :: force_id, mtilde_id
  integer :: tmp_aid, tmp_fid, tmp_mid1, tmp_mid2, nth

  tmp_mid1 = mtilde_id(kappa1, csw1, rho1, theta, nf)
  tmp_mid2 = mtilde_id(kappa2, csw2, rho2, theta, nf)

  tmp_aid = 4
  tmp_fid = force_id(type, tmp_mid1)

  n_fermi_id = n_fermi_id + 1

  if(n_fermi_id > size(action%fermi)) then
     write(0,*) "n_fermi_id > size(action%fermi) :: size is",size(action%fermi)
     stop
  endif

  nth=0
  if ( type_rational(type) ) then

     if (type == EO_RATIONAL) then
        if (nf==2) then
           nth=input%hmc_mpf_rhmc_l
        else
           nth=input%hmc_mpf_rhmc_s * 2
        endif
     endif
     if (type == DD_RATIONAL) then
        if (nf==2) then
           nth=input%hmc_mpf_dd_rhmc_l
        else
           nth=input%hmc_mpf_dd_rhmc_s * 2
        endif
     endif
     if (type == HH_RATIONAL) then
        if (nf==2) then
           nth=input%hmc_mpf_hh_rhmc_l
        else
           nth=input%hmc_mpf_hh_rhmc_s * 2
        endif
     endif
     if (type == UN_RATIONAL) then
        if (nf==2) then
           nth=input%hmc_mpf_un_rhmc_l
        else
           nth=input%hmc_mpf_un_rhmc_s * 2
        endif
     endif
  endif

  action%fermi(n_fermi_id)%aid   = tmp_aid
  action%fermi(n_fermi_id)%fid   = tmp_fid
  action%fermi(n_fermi_id)%mid1  = tmp_mid1
  action%fermi(n_fermi_id)%mid2  = tmp_mid2
  action%fermi(n_fermi_id)%type  = type
  action%fermi(n_fermi_id)%nth   = nth
  if (nth==0) then
     action%fermi(n_fermi_id)%rid = -999
  else
     action%fermi(n_fermi_id)%rid = 0
  endif

end

!-------------------------------------------------------------------------------
logical function type_rational(type)
  implicit none
  integer, intent(in) :: type

  type_rational=.false.
  type_rational=( type == EO_RATIONAL .or. &
                 type == UN_RATIONAL .or. &
                 type == DD_RATIONAL .or. &
                 type == HH_RATIONAL )

end
!===============================================================================
