!===============================================================================
!
!  init_imp.F90 - init values for action
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_action(para)
  use module_action
  use module_switches
  use typedef_para
  implicit none

  type(type_para), intent(in) :: para
  integer :: n_term, n_gauge, n_clover, n_logdet, n_other, nf, n_fermi
  integer :: aid, fid, cid, sid

  n_term = 0
  n_gauge = 0
  n_logdet = 0
  aid = 0
  fid = 0
  cid = 0
  sid = 0

  nf = 0
  n_fermi = 0
  n_clover = 0

  n_other = 0

  if (gauge_action /= "NON") then
     n_term = n_term + 1
     n_gauge = n_gauge + 1
     aid = aid + 1
     fid = fid + 1
     action%plaq%aid = aid
     action%plaq%fid = fid
     action%plaq%sid = sid
  endif

  if (gauge_action /= "WILSON") then
     n_term = n_term + 1
     n_gauge = n_gauge + 1
     aid = aid + 1
     fid = fid + 1
     action%impg%aid = aid
     action%impg%fid = fid
     action%impg%sid = sid
 endif
end

!-------------------------------------------------------------------------------
subroutine init_imp_all(doall)
  use module_stout_field
  use module_field
  use module_action
  use module_function_decl
  use module_switches
  implicit none
  logical :: doall
  integer :: i
  integer, save :: count=0

  DEBUG2S("Start: init_imp_all")

  if (doall) then
     gauge(1:size(gauge))%init = .true.
     if (associated(stout)) stout(1:size(stout))%init = .true.
     if (associated(clover))clover(1:size(clover))%init = .true.
  endif

  if (associated(stout))then
     do i = 1, size(stout)
        call stout_link_smearing(i)
     enddo
  endif
  call init_gauge(2)
  if (associated(clover))then
     do i = 1, size(clover)
        call init_clover(i)
     enddo
  endif
  !
  ! check
  !
  if(my_pe()==0 .and. count==0 ) then

  write(0,*)"================================================================="
  write(0,*)"=======================================================check para"

  if (associated(action%fermi))then
     write(0,*)"%fermi: aid is defined at other place, and "
     write(0,*)"%fermi: does not print more than 3 fermionic action"
     write(0,100)"%fermi","ID", "aid", "fid", "mid1", "mid2", "type", " rid", " nth"
     do i = 1, size(action%fermi)
        write(0,101)"%fermi", i, &
                 action%fermi(i)%aid, action%fermi(i)%fid, &
                 action%fermi(i)%mid1,action%fermi(i)%mid2,&
                 action%fermi(i)%type, &
                 action%fermi(i)%rid, action%fermi(i)%nth
     enddo
  endif

  if (associated(action%mtilde))then
     write(0,102)"%mtilde","ID", "cid", "did", "rho", "kappa", "cswk","h","b"
     do i = 1, n_mtilde_id
        write(0,103)"%mtilde", i, &
        action%mtilde(i)%cid, action%mtilde(i)%did, &
        action%mtilde(i)%rho, action%mtilde(i)%kappa, &
        action%mtilde(i)%cswkappa, &
        action%mtilde(i)%h, action%mtilde(i)%b
     enddo
  endif

  if (associated(action%logdet%cid))then
     write(0,104)"%logdet","cid", "aid", "fid"
     do i = 1, size(action%logdet%cid)
        write(0,105)"%logdet",action%logdet%cid(i), &
                              action%logdet%aid, &
                              action%logdet%fid
     enddo
  endif

  if (associated(clover))then
     write(0,106)"%clover","ID", "nf", "csw","kappa","cswk","theta"
     do i = 1, size(clover)
        write(0,107)"%clover",i,clover(i)%nf, clover(i)%csw, &
                    clover(i)%kappa, clover(i)%cswkappa, clover(i)%theta
     enddo
  endif

  endif !! of if (my_pe()==0)
  if (.not. (gauge_action == "TREETAD" .or. switches%tadcsw)) count=1

  DEBUG2S("End: init_imp_all")
100 format (a7,3a4,3a5, 2a6)
101 format (a7,3i4,3i5, 2i6)
102 format (a7,3a4,a9  ,a9  ,a18   ,a9  ,a20   )
103 format (a7,3i4,f9.6,f9.6,f18.15,f9.6,f20.15)
104 format (a7,3a4)
105 format (a7,3i4)
106 format (a7,2a4,a9  ,a9  ,a18   ,a20   )
107 format (a7,2i4,f9.6,f9.6,f18.15,f20.15)
end

!-------------------------------------------------------------------------------
subroutine init_quark(mid)
  use module_action
  implicit none
  integer, intent(in) :: mid

  if (mid == 0) return
  if (action%mtilde(mid)%cid /= 0)call init_clover(action%mtilde(mid)%cid)
  if (action%mtilde(mid)%did /= 0)call init_gauge(2)
end

!-------------------------------------------------------------------------------
subroutine init_gauge(id)
  use module_stout_field
  use module_field
  use module_function_decl
  use module_switches
  implicit none
  integer, intent(in) :: id

  if (id == 2 .and. (.not. switches%dynamical)) return
  if (id == 3 .and. (.not. switches%clover)) return
  if (.not. gauge(id)%init) return

  DEBUG2S("Start: init_gauge")

  if (gauge(id)%sid /= 0) then
     call stout_link_smearing(1)
     if (id == 2) then 
        call flip_bc2(gauge(id)%u,  stout(1)%iter(stout(1)%n)%u)
        call gauge_copy_r8_to_r4(gauge_r4(id)%u, gauge(id)%u)
     endif
     if (id == 3) gauge(id)%u =>  stout(1)%iter(stout(1)%n)%u
  else
     if (id == 2) then
        call flip_bc2(gauge(id)%u,  gauge(1)%u)
        call gauge_copy_r8_to_r4(gauge_r4(id)%u, gauge(id)%u)
     endif
     if (id == 3) gauge(id)%u =>  gauge(1)%u
  endif

!!#ifdef LIBDI
!!  if (id == 2) call d_g_u_reorder(gauge(2)%u)
!!  if (id == 2) call d_g_init(gauge(2)%u)
!!#endif
!!#ifdef BAGEL
!!  if (id == 2) call copy_u_bagel_d()
!!#endif

  if (id == 2) then
     select case (version_of_d())
        case(4,45,8,85,520,100)
           call d_g_init(gauge(2)%u)
     end select
     call u2ddu(gauge(2)%u)
#ifdef BAGEL
     call copy_u_bagel_d()
#endif
  endif

  gauge(id)%init = .false.
  DEBUG2S("End: init_gauge")
end

!-------------------------------------------------------------------------------
subroutine init_clover(id)
  use module_field
  use module_switches
  implicit none
  integer,intent(in) :: id
  REAL :: m1, m2

  if (.not. switches%clover) return
  if (.not. clover(id)%init) return

  DEBUG2S("Start: init_clover")
  call init_gauge(3)
  DEBUG2S("Start: init_clover 1")
  call theta2m1m2(clover(id)%theta, m1, m2)
  DEBUG2S("Start: init_clover 2")

!!  write(0,*)id, size(clover), size(gauge)
!!  stop

  call clover_init(clover(id)%a, clover(id)%i, clover(id)%b, gauge(1)%u, &
                   clover(id)%cswkappa, m1, m2)
  DEBUG2S("Start: init_clover 3")
  call clover_a_copy_r8_to_r4(clover_r4(id)%a, clover(id)%a)
  call clover_a_copy_r8_to_r4(clover_r4(id)%i, clover(id)%i)
  call clover_b_copy_r8_to_r4(clover_r4(id)%b, clover(id)%b)

  DEBUG2S("Start: init_clover 4")
  call clover2cloverdd(id)

  clover(id)%init = .false.
  DEBUG2S("End: init_clover")
end
!-------------------------------------------------------------------------------
subroutine theta2m1m2(theta, m1, m2)
  implicit none
  REAL,intent(out) :: m1, m2
  REAL,intent(in)  :: theta

  m1 = ZERO
  m2 = ZERO
  if (theta == ZERO) return
  m1 =   theta !! / dble(nf)
  m2 = - theta !! / dble(nf)
!!     m1 = m1 * (ONE - para%kappa /0.138984_8)
!!     m2 = m2 * (ONE - para%kappa /0.138984_8)
!!     m1 = cosh(para%theta / TWO) - ONE + sinh(para%theta / TWO)
!!     m2 = cosh(para%theta / TWO) - ONE - sinh(para%theta / TWO)
!!     m1 = m1 * (ONE - para%kappa * EIGHT)
!!     m2 = m2 * (ONE - para%kappa * EIGHT)

end

!===============================================================================
